"""
Config module: apps config data model and json loader
"""
from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Optional, Type, Union, List, Generic

from hopeit.dataobjects import EventPayloadType, dataobject
from hopeit.dataobjects.payload import Payload
from hopeit.server.config import replace_config_args, replace_env_vars, ServerConfig, AuthType
from hopeit.server.names import auto_path

__all__ = ['AppDescriptor',
           'AppSettings',
           'Env',
           'EventType',
           'EventPlugMode',
           'EventDescriptor',
           'EventSettings',
           'ReadStreamDescriptor',
           'WriteStreamDescriptor',
           'EventLoggingConfig',
           'EventStreamConfig',
           'Compression',
           'Serialization',
           'AppEngineConfig',
           'AppConfig',
           'parse_app_config_json']


@dataobject
@dataclass
class AppDescriptor:
    """
    App descriptor
    """
    name: str
    version: str

    def __post_init__(self):
        if len(self.name) == 0:
            raise ValueError('name', self)
        if len(self.version) == 0:
            raise ValueError('version', self)

    def app_key(self):
        return auto_path(self.name, self.version)


Env = Dict[str, Dict[str, Union[str, float, bool]]]
AppSettings = Dict[str, Dict[str, Any]]


class EventType(str, Enum):
    """
    Supported event types

    GET: event triggered from api get endpoint
    POST: event triggered from api post endpoint
    STREAM: event triggered read events from stream. Can be started and stopped.
    SERVICE: event executed on demand or continuously. Long lived. Can be started and stopped.
    MULTIPART: event triggered from api postform-multipart request via endpoint.
    """
    GET = 'GET'
    POST = 'POST'
    STREAM = 'STREAM'
    SERVICE = 'SERVICE'
    MULTIPART = 'MULTIPART'


class StreamQueue:
    AUTO = "AUTO"

    @classmethod
    def default_queues(cls):
        return [cls.AUTO]


@dataobject
@dataclass
class ReadStreamDescriptor:
    """
    Configuration to read streams

    :field stream_name: str, base stream name to read
    :consumer_group: str, consumer group to send to stream processing engine to keep track of
        next messag to consume
    :queues: List[str], list of queue names to poll from. Each queue act as separate stream
        with queue name used as stream name suffix, where `AUTO` queue name means to consume
        events when no queue where specified at publish time, allowing to consume message with different
        priorities without waiting for all events in the stream to be consumed.
        Queues specified in this entry will be consumed by this event
        on each poll cycle, on the order specified. If not present
        only AUTO queue will be consumed. Take into account that in applications using multiple
        queue names, in order to ensure all messages are consumed, all queue names should be listed
        here including AUTO, except that the app is intentionally designed for certain events to
        consume only from specific queues. This configuration is manual to allow consuming messages
        produced by external apps.
    """
    name: str
    consumer_group: str
    queues: List[str] = field(default_factory=StreamQueue.default_queues)


class StreamQueueStrategy(str, Enum):
    """
    Different strategies to be used when reading streams from a queue and writing to another stream.

    :field PROPAGATE: original queue name will be preserved, so messages consumed from a queue will
        maintain that queue name when published
    :field DROP: queue name will be dropped, so messages will be published only to queue specified in
        `write_stream` configuration, or default queue if not specified.
    """
    PROPAGATE = "PROPAGATE"
    DROP = "DROP"


@dataobject
@dataclass
class WriteStreamDescriptor:
    """
    Configuration to publish messages to a stream

    :field: name, str: stream name
    :field: queue, List[str], queue names to be used to publish to stream.
        Each queue act as separate stream with queue name used as stream name suffix,
        allowing to publish messages to i.e. a queue that will be consumed with priority,
        or to multiple queues that will be consumed by different readers.
        Queue suffix will be propagated through events, allowing an event in a defined queue
        and successive events in following steps to be consumed using same queue name.
        Notice that queue will be applied only to messages coming from default queue
        (where queue is not specified at intial message creation). Messages consumed
        from other queues will be published using same queue name as they have when consumed.
    :field queue_stategory: strategy to be used when consuming messages from a stream
        with a queue name and publishing to another stream. Default is `StreamQueueStrategy.DROP`,
        so in case of complex stream propagating queue names are configured,
        `StreamQueueStrategy.PROPAGATE` must be explicitly specified.
    """
    name: str
    queues: List[str] = field(default_factory=StreamQueue.default_queues)
    queue_strategy: StreamQueueStrategy = StreamQueueStrategy.DROP


@dataobject
@dataclass
class EventLoggingConfig:
    """
    Logging configuration specific for the event

    :field extra_fields: list of str, extra fields required to apps when logging
        as part of extra(...) call
    :field stream_fields: list of str, field names to extra when reading streams,
        valid options are
            'name': stream name,
            'msg_id', internal message id
            'consumer_group', conaumer group name
            'submit_ts', utc time message was submited to stream
            'event_ts', event timestamp from @data_event
            'event_id', event id from @data_event
            'read_ts': uct time when message was consumed from stream
    """
    extra_fields: List[str] = field(default_factory=list)
    stream_fields: List[str] = field(default_factory=list)

    def __post_init__(self):
        if len(self.stream_fields) == 0:
            self.stream_fields = ['name', 'msg_id', 'consumer_group']
        self.stream_fields = [k if k.startswith('stream.') else f"stream.{k}"
                              for k in self.stream_fields]


class Compression(str, Enum):
    """
    Available compression algorithms and levels for event payloads.
    """
    NONE = 'none'
    LZ4 = 'lz4'
    LZ4_MIN = 'lz4:0'
    LZ4_MAX = 'lz4:16'
    ZIP = 'zip'
    ZIP_MIN = 'zip:1'
    ZIP_MAX = 'zip:9'
    GZIP = 'gzip'
    GZIP_MIN = 'gzip:1'
    GZIP_MAX = 'gzip:9'
    BZ2 = 'bz2'
    BZ2_MIN = 'bz2:1'
    BZ2_MAX = 'bz2:9'
    LZMA = 'lzma'


class Serialization(str, Enum):
    """
    Available serialization methods for event payloads.
    """
    JSON_UTF8 = 'json'
    JSON_BASE64 = 'json+base64'
    PICKLE3 = 'pickle:3'
    PICKLE4 = 'pickle:4'
    PICKLE5 = 'pickle:5'


@dataobject
@dataclass
class EventStreamConfig:
    """
    Stream configuration for STREAM events
    :field timeout: float, timeout for stream processing im seconds. If timeout is exceeded event
        processing will be cancelled. Default 60 seconds
    :field target_max_len: int, default 0, max number of elements to be used as a target
        for the stream collection size. Messages above this value might be dropped
        from the collection when new items are added. Notice that the number of the items
        in the collection could exceed temporary this value to allow optimized behaviour,
        but no items will be dropped until the collection exceeds target_max_len.
        With default value of 0, collection size is unlimited and items should be removed by apps.
    :field throttle_ms: int, milliseconds specifying minimum duration for each event
    :filed step_delay: int, milliseconds to sleep between steps
    :field batch_size: int, number of max messages to process each time when reading stream,
        set to 1 to ensure min losses in case of process stop, set higher for performance
    :field compression: Compression, compression algorithm used to send messages to stream, if not specified
        default from Server config will be used.
    :field serialization: Serialization, serialization method used to send messages to stream, if not specified
        default from Server config will be used.
    """
    timeout: float = 60.0
    target_max_len: int = 0
    throttle_ms: int = 0
    step_delay: int = 0
    batch_size: int = 100
    compression: Optional[Compression] = None
    serialization: Optional[Serialization] = None


@dataobject
@dataclass
class EventSettings(Generic[EventPayloadType]):
    """
    Event execution configuration, used to parse `settings` entries in app config file where the key
    is the event name.

    :field response_timeout: float, default 60.0: seconds to timeout waiting for event execution
        when invoked externally .i.e. GET or POST events. If exceeded, Timed Out response will be returned.
        Notice that this timeout does not apply for stream processing events. Use EventStreamsConfig.timeout
        to set up timeout on stream processing.
    :field logging: EventLoggingConfig, configuration for logging for this particular event
    :field stream: EventStreamConfig, configuration for stream processing for this particular event
    """
    response_timeout: float = 60.0
    logging: EventLoggingConfig = field(default_factory=EventLoggingConfig)
    stream: EventStreamConfig = field(default_factory=EventStreamConfig)
    extras: Dict[str, Any] = field(default_factory=dict)

    def __call__(self, *, key: str = '_', datatype: Type[EventPayloadType]) -> EventPayloadType:
        return Payload.from_obj(self.extras.get(key, {}), datatype=datatype)


class EventPlugMode(str, Enum):
    """
    Defines how an event route is plugged into apps when
    it is used as a plugin.

    STANDALONE: The event is added as a normal route where it is defined. Not added to apps.
    ON_APP: The event route is added only to app routes where it is used as a plugin.
    """
    STANDALONE = 'Standalone'
    ON_APP = 'OnApp'


class EventConnectionType(str, Enum):
    """
    Event connection type
    """
    GET = "GET"
    POST = "POST"


@dataobject
@dataclass
class EventConnection:
    """
    EventConnection: describes dependencies on this event when calling
    event on apps configured in `app_connections` sections. Only events
    specified are allowed to be invoked using `hopeit.client`

    :field: app_connection, str: key of app entry used in app_connections sections
    :field: event, str: target event_name to be called
    :filed: type, EventConnectionType: a valid event connection type, i.e. GET or POST
    :field: route, optional str: custom route in case event is not attached to default `app/version/event`
    """
    app_connection: str
    event: str
    type: EventConnectionType


@dataobject
@dataclass
class EventDescriptor:
    """
    Event Descriptor: configures event implementation

    :field: type, EventType: type of event i.e.: GET, POST, MULTIPART, STREAM, SERVICE
    :field: plug_mode, EventPlugMode: defines whether an event defined in a plugin is created in the
        current app (ON_APP) or it will be created in the original plugin (STANDALONE, default)
    :field: route, optional str: custom route for endpoint. If not specified route will be derived
        from `/api/app_name/app_version/event_name`
    :field: impl, optional str: custom event implementation Python module. If not specified, module
        with same same as event will be imported.
    :field: connections, list of EventConnection: specifies dependencies on other apps/endpoints,
        that can be used by client plugins to call events on external apps
    :field: read_stream, optional ReadStreamDescriptor: specifies source stream to read from.
        Valid only for STREAM events.
    :field: write_stream, optional WriteStreamDescriptor: for any type of events, resultant dataobjects will
        be published to the specified stream.
    :field: auth, list of AuthType: supported authentication schemas for this event. If not specified
        application default will be used.
    :field: setting_keys, list of str: by default EventContext will have access to the settings section
        with the same name of the event using `settings = context.settings(datatype=MySettingsType)`.
        In case additional sections are needed to be accessed from
        EventContext, then a list of setting keys, including the name of the event if needed,
        can be specified here. Then access to a `custom` key can be done using
        `custom_settings = context.settings(key="customer", datatype=MyCustomSettingsType)`
    :field: dataobjects, list of str: list of full qualified dataobject types that this event can process.
        When not specified, the engine will inspect the module implementation and find all datatypes supported
        as payload in the functions defined as `__steps__`. In case of generic functions that support
        `payload: DataObject` argument, then a list of full qualified datatypes must be specified here.
    :field: group, str: group name, if none is assigned it is automatically assigned as 'DEFAULT'.
    """
    DEFAULT_GROUP = 'DEFAULT'

    type: EventType
    plug_mode: EventPlugMode = EventPlugMode.STANDALONE
    route: Optional[str] = None
    impl: Optional[str] = None
    connections: List[EventConnection] = field(default_factory=list)
    read_stream: Optional[ReadStreamDescriptor] = None
    write_stream: Optional[WriteStreamDescriptor] = None
    auth: List[AuthType] = field(default_factory=list)
    setting_keys: List[str] = field(default_factory=list)
    dataobjects: List[str] = field(default_factory=list)
    group: str = DEFAULT_GROUP

    def __post_init__(self):
        if self.read_stream:
            assert '{auto}' not in self.read_stream.name, \
                "read_stream.name should be defined. {auto} is not allowed."


@dataobject
@dataclass
class AppEngineConfig:
    """
    Engine specific parameters shared among events

    :field import_modules: list of string with the python module names to import to find
        events and datatype implementations
    :field read_stream_timeout: timeout in milliseconds to block connection pool when waiting for stream events
    :field read_stream_interval: delay in milliseconds to wait before attempting a new batch. Use to prevent
        connection pool to be blocked constantly.
    :track_headers: list of required X-Track-* headers
    :cors_origin: allowed CORS origin for web server
    """
    import_modules: Optional[List[str]] = None
    read_stream_timeout: int = 1000
    read_stream_interval: int = 1000
    default_stream_compression: Compression = Compression.LZ4
    default_stream_serialization: Serialization = Serialization.JSON_BASE64
    track_headers: List[str] = field(default_factory=list)
    cors_origin: Optional[str] = None

    def __post_init__(self):
        self.track_headers = [k if k.startswith('track.') else f"track.{k}" for k in self.track_headers]
        if 'track.request_ts' not in self.track_headers:
            self.track_headers = ['track.request_ts'] + self.track_headers
        if 'track.request_id' not in self.track_headers:
            self.track_headers = ['track.request_id'] + self.track_headers


@dataobject
@dataclass
class AppConnection:
    """
    AppConnections: metadata to initialize app client in order to connect
    and issue requests to other running apps

    :field: name, str: target app name to connect to
    :field: version, str: target app version
    :field: client, str: hopeit.app.client.Client class implementation, from available client plugins
    :field: settings, optional str: key under `settings` section of app config containing connection configuration,
        if not specified, plugin will lookup its default section usually the plugin name. But in case multiple
        clients need to be configured, this value can be overridden.
    """
    name: str
    version: str
    client: str = "<<NO CLIENT CONFIGURED>>"
    settings: Optional[str] = None
    plugin_name: Optional[str] = None
    plugin_version: Optional[str] = None


@dataobject
@dataclass
class AppConfig:
    """
    App Configuration container
    """
    app: AppDescriptor
    engine: AppEngineConfig = field(default_factory=AppEngineConfig)
    app_connections: Dict[str, AppConnection] = field(default_factory=dict)
    env: Env = field(default_factory=dict)
    events: Dict[str, EventDescriptor] = field(default_factory=dict)
    server: Optional[ServerConfig] = None
    plugins: List[AppDescriptor] = field(default_factory=list)
    settings: AppSettings = field(default_factory=dict)
    effective_settings: Optional[AppSettings] = None

    def app_key(self):
        return self.app.app_key()

    def setup(self):
        self.effective_settings = {}
        self._setup_streams_settings()
        self._setup_event_extra_settings()
        return self

    def _setup_streams_settings(self):
        """
        Prepares missing or default values for events stream configuration:

        1) Completes missing event streams settings using settings key
            from write_stream or read_stream name
        2) Sets default compression and serialization for streams based on server config
        """
        for event_name, event_info in self.events.items():
            stream_settings = {}
            if event_info.write_stream:
                stream_settings.update(self.settings.get(event_info.write_stream.name, {}))
            elif event_info.read_stream:
                stream_settings.update(self.settings.get(event_info.read_stream.name, {}))
            settings_data = deepcopy(self.settings.get(event_name, {"stream": {}}))
            stream_settings.update(settings_data.get('stream', {}))
            settings_data['stream'] = stream_settings
            settings = EventSettings.from_dict(settings_data, validate=True)
            if settings.stream.compression is None:
                settings.stream.compression = self.engine.default_stream_compression
            if settings.stream.serialization is None:
                settings.stream.serialization = self.engine.default_stream_serialization
            self.effective_settings[event_name] = settings.to_dict()

    def _setup_event_extra_settings(self):
        """
        Add to event settings section extra keys references in `setting_keys` property
        of EventDescriptor, so these are available in `EventContext` using
        `context.settings(key, datatype)

        For these extra keys, validation is not perform until accessed.
        """
        for event_name, event_info in self.events.items():
            extras = {}
            if event_name in self.settings:
                extras['_'] = self.settings[event_name]
            for key in event_info.setting_keys:
                extras[key] = self.settings[key]
            self.effective_settings[event_name]['extras'] = extras


def parse_app_config_json(config_json: str) -> AppConfig:
    """
    Parses configuration json file contents into AppConfig data structure.
    Before conversion, parameters enclosed with { } are replaced by its
    respective values (@see _replace_args)
    """
    # effective_config_json = _replace_args(config_json)
    effective_config_json = replace_env_vars(config_json)
    app_config = AppConfig.from_json(effective_config_json)  # type: ignore
    replace_config_args(
        parsed_config=app_config,
        config_classes=(
            AppDescriptor, EventDescriptor, ReadStreamDescriptor, WriteStreamDescriptor,
            AppConnection, EventConnection, EventSettings
        ),
        auto_prefix=app_config.app.app_key()
    )
    return app_config.setup()



"""
Payload tools to serialize and deserialze event payloads and responses, including dataobjects
"""

import json
from typing import Type, Generic, Optional, Union

from dataclasses_jsonschema import ValidationError

from hopeit.dataobjects import EventPayloadType

_ATOMIC_TYPES = (str, int, float, bool)
_COLLECTION_TYPES = (dict, list, set)
_MAPPING_TYPES = (dict, )
_LIST_TYPES = (list, set)
_UNORDERED_LIST_TYPES = (set, )


class Payload(Generic[EventPayloadType]):
    """
    Convenience ser/deser functions for @dataobject decorated object (@see DataObject)
    """
    @staticmethod
    def from_json(json_str: Union[str, bytes],
                  datatype: Type[EventPayloadType],
                  key: str = 'value') -> EventPayloadType:
        """
        Converts json_str to desired datatype

        :param json_str: str containing valid json,
            or string representation for atomic values
        :param datatype: supported types defined in EventPayload
        :param key: key to extract atomic types from
        :return: instance of datatype
        """
        if datatype in _ATOMIC_TYPES:
            return datatype(json.loads(json_str).get(key))  # type: ignore
        if datatype in _COLLECTION_TYPES:
            return datatype(json.loads(json_str))  # type: ignore
        try:
            return datatype.from_json(json_str, validate=datatype.__data_object__['validate'])  # type: ignore
        except ValidationError as e:
            raise ValueError(f"Cannot read JSON: type={datatype} validation_error={str(e)}") from e
        except Exception:
            assert getattr(datatype, 'from_json'), \
                f"{datatype} should be annotated with @dataobject"
            raise  # Raises unexpected exceptions, if assert block does not catch missing @dataobject

    @staticmethod
    def from_obj(data: Union[dict, list],
                 datatype: Type[EventPayloadType],
                 key: str = 'value',
                 item_datatype: Optional[Type[EventPayloadType]] = None) -> EventPayloadType:
        """
        Converts dictionary to desired datatype

        :param data: dictionary containing fields expected on datatype
        :param datatype: supported types defined in EventPayload
        :param key: key to extract atomic types from
        :param item_datatype: optional datatype to parse items in collections
        :return: instance of datatype
        """
        if datatype in _ATOMIC_TYPES:
            return datatype(data.get(key))  # type: ignore
        if datatype in _MAPPING_TYPES:
            if item_datatype and isinstance(data, _MAPPING_TYPES):
                return {  # type: ignore
                    k: Payload.from_obj(v, item_datatype, key) for k, v in data.items()
                }
            return datatype(data)  # type: ignore
        if datatype in _LIST_TYPES:
            if item_datatype and isinstance(data, _LIST_TYPES):
                return datatype([  # type: ignore
                    Payload.from_obj(v, item_datatype, key) for v in data
                ])
            return datatype(data)  # type: ignore
        assert getattr(datatype, 'from_dict'), \
            f"{datatype} should be annotated with @dataobject"
        try:
            return datatype.from_dict(data, validate=datatype.__data_object__['validate'])  # type: ignore
        except ValidationError as e:
            raise ValueError(f"Cannot read object: type={datatype} validation_error={str(e)}") from e

    @staticmethod
    def to_json(payload: EventPayloadType, key: Optional[str] = 'value') -> str:
        """
        Converts event payload to json string

        :param payload: EventPayload, instance of supported object type
        :param key: key name used in generated json when serializing atomic values
        :return: str containing json representation of data. In case of simple datatypes,
            a json str of key:value form will be generated using key parameter if it's not None.
        """
        if isinstance(payload, _ATOMIC_TYPES):  # immutable supported types
            if key is None:
                return json.dumps(payload)
            return json.dumps({key: payload})
        if isinstance(payload, _LIST_TYPES):
            return "[" + ', '.join(Payload.to_json(item, key=None) for item in payload) + "]"
        if isinstance(payload, _MAPPING_TYPES):
            return "{" + ', '.join(
                f'"{str(k)}": {Payload.to_json(item, key=None)}' for k, item in payload.items()
            ) + "}"
        assert getattr(payload, 'to_json'), \
            f"{type(payload)} should be annotated with @dataobject"
        try:
            return payload.to_json(validate=payload.__data_object__['validate'])  # type: ignore
        except (ValidationError, AttributeError) as e:
            raise ValueError(f"Cannot convert to JSON: type={type(payload)} validation_error={str(e)}") from e

    @staticmethod
    def to_obj(payload: EventPayloadType, key: Optional[str] = 'value') -> Union[dict, list]:
        """
        Converts event payload to dictionary or list

        :param payload: EventPayload, instance of supported object type
        :param key: key name used in generated json when serializing atomic values
        :return: dict or list containing mapped representation of data. In case of simple datatypes,
            a key:value form will be generated using key parameter. All objects mappable to dict will
            be converted. Flat collections will be converted to list.
        """
        if isinstance(payload, _ATOMIC_TYPES):  # immutable supported types
            if key is None:
                return payload  # type: ignore  # only for recursive use
            return {key: payload}
        if isinstance(payload, _UNORDERED_LIST_TYPES):
            return [Payload.to_obj(v, key=None) for v in sorted(payload)]
        if isinstance(payload, _LIST_TYPES):
            return [Payload.to_obj(v, key=None) for v in payload]
        if isinstance(payload, _MAPPING_TYPES):
            return {k: Payload.to_obj(v, key=None) for k, v in payload.items()}
        assert getattr(payload, 'to_dict'), \
            f"{type(payload)} should be annotated with @dataobject"
        try:
            return payload.to_dict(validate=payload.__data_object__['validate'])  # type: ignore
        except (ValidationError, AttributeError) as e:
            raise ValueError(f"Cannot convert to dict: type={type(payload)} validation_error={str(e)}") from e

    @staticmethod
    def parse_form_field(field_data: Union[str, dict], datatype: Type[EventPayloadType],
                         key: str = 'value') -> EventPayloadType:
        """Helper to parse dataobjects from form-fields where encoding type is not correctly set to json"""
        if isinstance(field_data, str):
            return Payload.from_json(field_data, datatype, key)
        return datatype.from_dict(field_data)  # type: ignore



"""
hopeit engine server modules

Modules Overview:

    * **api**: provides openapi support for endpoints defined in web module.
    * **web**: provides /api endpoints creation for GET and POST events, and provides
      /mgmt endpoints for STREAM and SERVICE events
    * **engine**: provides App initialization and containers
    * **events**: provides handlers for execution of App Events
    * **streams**: provides read-write support for Redis Streams
    * **steps**: helpers to execute event steps
    * **config**: helpers to load configuration files
    * **imports**: helpers to import App Events modules
    * **logger**: provides base logging for server
    * **metrics**: provides metrics calculation for events
    * **names**: helpers for events and route naming conventions
    * **errors**: helpers for error messages
"""
__all__ = ['config',
           'engine',
           'errors',
           'events',
           'imports',
           'logger',
           'metrics',
           'names',
           'steps',
           'web',
           'api']



"""
CLI openapi commands
"""
import sys
import re

try:
    import click
    from deepdiff import DeepDiff  # type: ignore

    from hopeit.app.config import parse_app_config_json
    from hopeit.server import api
    from hopeit.server.config import parse_server_config_json
    from hopeit.server.logger import engine_logger
except ModuleNotFoundError:
    print("ERROR: Missing dependencies."
          "\n       To use hopeit_server command line tool"
          "\n       install using `pip install hopeit.engine[web,cli]`")
    sys.exit(1)

logger = engine_logger().init_cli('openapi')
setattr(api, 'logger', logger)


@click.group()
def openapi():
    pass


@openapi.command()
@click.option('--api-version', prompt='API Version', help='API Version string x.x.x.')
@click.option('--title', prompt='API Title', help='API title string.')
@click.option('--description', prompt='API Description', help='API description string.')
@click.option('--config-files', help='Comma-separated list of server, plugins and app config files.')
@click.option('--output-file', help='Path to json api file to be loaded.')
@click.option('--generate', is_flag=True, default=False, help='Indicates to generate paths for all events.')
def create(api_version: str, title: str, description: str, config_files: str,
           output_file: str, generate: bool):
    """
    Creates OpenAPI spec file
    """
    logger.info(__name__, "openapi.update")
    if generate:
        api.setup(generate_mode=True)
    api.init_empty_spec(api_version, title, description)
    _update_api_spec(config_files)
    api.static_spec = api.spec
    api.save_api_file(output_file, api_version)


@openapi.command()
@click.option('--api-version',
              help='API Version string x.x.x. Needs to be incremented if there are spec changes during update.')
@click.option('--config-files', help='Comma-separated list of server, plugins and app config files.')
@click.option('--input-file', help='Path to json api file to be loaded.')
@click.option('--output-file', help='Path to json api file to be loaded.')
@click.option('--generate', is_flag=True, default=False, help='Indicates to generate paths for all events.')
def update(api_version: str, config_files: str, input_file: str, output_file: str, generate: bool = False):
    """
    Updates OpenAPI spec based on current input file plus running configuration from apps.
    """
    logger.info(__name__, "openapi.update")
    if generate:
        api.setup(generate_mode=True)
    api.load_api_file(input_file)
    _update_api_spec(config_files)
    api.save_api_file(output_file, api_version)


@openapi.command()
@click.option('--config-files', help='Comma-separated list of plugins and app config files.')
@click.option('--input-file', help='Path to json api file to be loaded.')
@click.option('--generate', is_flag=True, help='Indicates to generate paths for all events.')
def diff(config_files: str, input_file: str, generate: bool):
    """
    List differences between input OpenAPI spec file and computed API from apps.
    """
    logger.info(__name__, "openapi.diff")
    if generate:
        api.setup(generate_mode=True)
    api.load_api_file(input_file)
    _update_api_spec(config_files)
    ddiff = DeepDiff(api.static_spec, api.spec, ignore_order=True)
    if len(ddiff) > 0:
        logger.warning(__name__, "Running configuration differs from API spec. Check differences.")
        for reason, items in ddiff.items():
            print()
            print(reason)
            for item in items:
                print('\t', item)
                try:
                    old_value = str(eval(item, None, {'root': api.static_spec}))  # pylint: disable=eval-used
                except Exception as e:  # pylint: disable=broad-except
                    old_value = str(e)
                try:
                    new_value = str(eval(item, None, {'root': api.spec}))  # pylint: disable=eval-used
                except Exception as e:  # pylint: disable=broad-except
                    new_value = str(e)
                print('\t\t<<<', re.sub('\n', ' ', old_value))
                print('\t\t>>>', re.sub('\n', ' ', new_value))
    else:
        logger.info(__name__, "OK: API spec and running configuration match.")


def _update_api_spec(config_files: str):
    """
    Computes and updates api module OpenAPI spec from specified config files argument.
    """
    apps_config = []
    config_files_list = config_files.split(',')
    with open(config_files_list[0], 'r', encoding="utf-8") as sf:
        server_config = parse_server_config_json(sf.read())
        api.register_server_config(server_config)
    for path in config_files_list[1:]:
        with open(path, 'r', encoding="utf-8") as f:
            app_config = parse_app_config_json(f.read())
            apps_config.append(app_config)
    api.register_apps(apps_config)


cli = click.CommandCollection(sources=[openapi])


if __name__ == '__main__':
    cli()



"""
Helpers to provide uniform logging to engine and apps
"""
import logging
from functools import partial
from typing import Tuple, Callable

from hopeit.server.logger import extra_values

__all__ = ['app_logger',
           'app_extra_logger']

DEFAULT_APP_LOGGER = 'app_logger_default'


def app_logger() -> logging.Logger:
    """
    Returns a placeholder to allow engine to setup later proper logger
    Allows to reference `logger` as a module variable.

    Use at module level in events implementation::

        from hopeit.logger import app_logger()

        logger = app_logger()

    """
    return logging.getLogger(DEFAULT_APP_LOGGER)


def app_extra_logger() -> Tuple[logging.Logger, Callable]:
    """
    Returns a placeholder to allow engine to setup later proper logger
    Allows to reference `logger` as a module variable,
    and the function `extra` to send extra values to logger.


    Use at module level in events implementation::

     from hopeit.logger import app_extra_logger()

     logger, extra = app_extra_logger()

    """
    return logging.getLogger(DEFAULT_APP_LOGGER), partial(extra_values, [])



"""
Provides generic `compress` and `decompress` methods for payloads
"""
from functools import partial
import zlib
import gzip
import bz2
import lzma
from typing import Callable, Tuple

import lz4.frame  # type: ignore

from hopeit.app.config import Compression

__all__ = ['compress', 'decompress']


def _compress_none(level: int, data: bytes) -> bytes:
    return data


def _decompress_none(data: bytes) -> bytes:
    return data


def _compress_lz4(level: int, data: bytes) -> bytes:
    return lz4.frame.compress(data, compression_level=level)


def _decompress_lz4(data: bytes) -> bytes:
    return lz4.frame.decompress(data)


def _compress_zlib(level: int, data: bytes) -> bytes:
    return zlib.compress(data, level=level)


def _decompress_zlib(data: bytes) -> bytes:
    return zlib.decompress(data)


def _compress_gzip(level: int, data: bytes) -> bytes:
    return gzip.compress(data, compresslevel=level)


def _decompress_gzip(data: bytes):
    return gzip.decompress(data)


def _compress_bz2(level: int, data: bytes) -> bytes:
    return bz2.compress(data, compresslevel=level)


def _decompress_bz2(data: bytes):
    return bz2.decompress(data)


def _compress_lzma(level: int, data: bytes) -> bytes:
    return lzma.compress(data)


def _decompress_lzma(data: bytes):
    return lzma.decompress(data)


_ALGOS = {
    'none': (_compress_none, 0, _decompress_none),
    'lz4': (_compress_lz4, 3, _decompress_lz4),
    'zip': (_compress_zlib, 6, _decompress_zlib),
    'gzip': (_compress_gzip, 9, _decompress_gzip),
    'bz2': (_compress_bz2, 9, _decompress_bz2),
    'lzma': (_compress_lzma, 0, _decompress_lzma)
}


def _compressors(compression: Compression) -> Tuple[Callable[[bytes], bytes], Callable[[bytes], bytes]]:
    info = compression.value.split(':')
    algo = info[0]
    comp, default_level, decomp = _ALGOS[algo]
    level = default_level if len(info) <= 1 else int(info[1])
    return partial(comp, level), decomp  # type: ignore


_COMPRESS = {
    c: _compressors(c)[0]
    for c in Compression
}

_DECOMPRESS = {
    c: _compressors(c)[1]
    for c in Compression
}


def compress(data: bytes, compression: Compression) -> bytes:
    return _COMPRESS[compression](data)


def decompress(data: bytes, compression: Compression) -> bytes:
    return _DECOMPRESS[compression](data)



"""
Data Objects type abstractions.
A Dataobject is basically any object wrapped in a Dataclass and that is able
to be stored, retrieved, serialized and deserialized by the platform toolkit.

Annotate dataclasses with @dataobject annotation to make it support:
    - JSON schema generation, validation and serialization/deserialization
    - Compression and Serialization using different mechanisms

Example:

    from hopeit.dataobjects import dataclass, dataobject

    @dataobject
    @dataclass
    class MyObject:
        name: str
        number: int
"""
import pickle
import uuid
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import TypeVar, Optional, Union, Any

from dataclasses_jsonschema import JsonSchemaMixin

__all__ = ['EventPayload',
           'EventPayloadType',
           'StreamEventParams',
           'dataobject',
           'DataObject',
           'copy_payload',
           'payload']


@dataclass
class StreamEventParams:
    """
    Helper class used to access attributes in @dataobject
    decorated objects, based on dot notation expressions
    """
    event_id_expr: Optional[str]
    event_ts_expr: Optional[str]

    @staticmethod
    def extract_attr(obj, expr):
        value = obj
        for attr_name in expr.split('.'):
            if value:
                value = getattr(value, attr_name)
        return value


class StreamEventMixin:
    """
    MixIn class to add functionality for DataObjects to provide event id and event timestamp
    to be suitable for stream publishing.

    Do not use this class directly, instead use `@databoject` class decorator.
    """

    def __init__(self, *, event_id_expr=None, event_ts_expr=None):
        self.__stream_event__ = StreamEventParams(event_id_expr, event_ts_expr)  # pragma: no cover
        raise NotImplementedError  # must use @dataobject decorator  # pragma: no cover

    def event_id(self) -> str:
        if self.__stream_event__.event_id_expr:
            return self.__stream_event__.extract_attr(self, self.__stream_event__.event_id_expr)
        return str(uuid.uuid4())

    def event_ts(self) -> Optional[datetime]:
        if self.__stream_event__.event_ts_expr:
            return self.__stream_event__.extract_attr(self, self.__stream_event__.event_ts_expr)
        return None


DataObject = TypeVar("DataObject", bound=JsonSchemaMixin)
EventPayload = Union[str, int, float, bool, dict, set, list, DataObject]
EventPayloadType = TypeVar("EventPayloadType")  # pylint: disable=invalid-name


class BinaryAttachment:
    """Type descriptor for multipart binary attachments"""


class BinaryDownload:
    """
    Type descriptor for binary download. Events returning files to download
    should specialize this class for specififc content types, i.e.:

    ```
    @dataobject
    @dataclass
    class ImagePng(BinaryDownload):
        content_type = "image/png"
    ```

    This way, the  type can be used in event API specification as response type.
    """
    content_type: str = "application/octet-stream"


def dataobject(
        decorated_class=None, *,
        event_id: Optional[str] = None,
        event_ts: Optional[str] = None,
        unsafe: bool = False,
        validate: bool = True,
        schema: bool = True):
    """
    Decorator for dataclasses intended to be used in API and/or streams. This decorated mainly implements
    JsonSchemaMixIn adding dataclass functionality to:
        * Generate Json Schema for Open API definition
        * Parse and convert from and to json
        * Validate Json against Json schema
        * Detect incompatibilities between API specification and payload structures
        * Detect undocumented API changes early (i.e. fields added or changed)

    In general, all dataclasses that are to be exchanged using API endpoints (payload and responses),
    or write and read from streams, need to implement @dataobject and decorator in addition to Python
    @dataclass decorator.

    In order to publish instances to streams, an event id and event timestamp
    can be extracted. StreamManager does that automatically for classes defining
    event_id() and event_ts() methods.

    This decorator, adds these two methods to a dataclass:
    `event_id()`: str, extract the id of the object from a given dot notation expression
    `event_ts()`: Optional[datetime], extract timestamp from a given dot notation expression

    :param decorated_class: decorated class
    :param event_id: optional str, dot notation expression
        to navigate to id field (i.e. 'id' or 'event.id')
    :param event_ts: optional str, dot notation expression
        to navigate to a datetime field (i.e. 'last_update.ts')
    :param unsafe: bool, default False. When False, every time a new step is invoked a copy of
        dataobject will be sent, preventing object to be mutated unintentionally.
        Specifying unsafe=True prevents making copies every time a step is invoked, improving performance.
        Use with caution: with unsafe=True, you can accidentally mutate objects after they are yield,
        specially when returning generators or `Spawn[...]`.
    :param validate: bool, default True: indicates whether to validate using JsonSchema automatically generated
        by `@dataobject` annotation when reading from and converting to json using `Json.from_json` and
        `Json.to_json`. Notice that if you call from_json or to_json directly from annotated dataobject
        you need to specify validate=True/False parameter value.
    :param schema: bool, default True: indicates to attempt json_schema generation in API module

    In case event_id is not provided, an uuid will be generated on each call to `event_id()`
    In case event_ts is not provided, None will be returned on each call to `event_ts()`

    Example::

        @dataobject
        @dataclass
        class StatusChange:
            ts: datetime
            status: str

        @dataobject(event_id='id', event_ts='last_status.ts', unsafe=True, validate=False)
        @dataclass
        class EventData:
            id: str
            last_status: StatusChange

    """

    def wrap(cls):
        amended_class = _add_jsonschema_support(cls)
        setattr(amended_class, '__data_object__', {'unsafe': unsafe, 'validate': validate, 'schema': schema})
        setattr(amended_class, '__stream_event__', StreamEventParams(event_id, event_ts))
        setattr(amended_class, 'event_id', StreamEventMixin.event_id)
        setattr(amended_class, 'event_ts', StreamEventMixin.event_ts)
        return amended_class

    if decorated_class is None:
        return wrap
    return wrap(decorated_class)


def _add_jsonschema_support(cls):
    if hasattr(cls, '__data_object__'):
        return cls
    if hasattr(cls, '__annotations__') and hasattr(cls, '__dataclass_fields__'):
        amended_class = type(cls.__name__,
                             (JsonSchemaMixin,) + cls.__mro__,
                             dict(cls.__dict__))
        return amended_class
    return cls


def _binary_copy(payload: Any) -> Any:
    return pickle.loads(pickle.dumps(payload, protocol=4))


def copy_payload(original: Optional[EventPayload]) -> Optional[EventPayload]:
    """
    Creates a copy of the original DataObject in case it is mutable.
    Returns original object in case it is a frozen dataclass
    """
    if original is None:
        return None
    if isinstance(original, (str, int, float, bool, tuple, Decimal)):  # immutable supported types
        return original
    if isinstance(original, (dict, set, list)):
        return _binary_copy(original)
    if hasattr(original, '__dataclass_params__') and original.__dataclass_params__.frozen:  # type: ignore
        return original
    if hasattr(original, '__data_object__') and original.__data_object__['unsafe']:
        return original
    return _binary_copy(original)



"""
Engine module: handle apps load, setup and serving
"""
import asyncio
import random
import uuid
from asyncio import CancelledError
from datetime import datetime, timezone
from typing import Awaitable, Optional, Dict, List, Union, Tuple, Any

from hopeit.server.imports import find_datobject_type, find_event_handler
from hopeit.server.steps import split_event_stages, event_and_step, extract_module_steps, \
    effective_steps
from hopeit.toolkit import auth
from hopeit.app.config import AppConfig, EventSettings, EventType, ReadStreamDescriptor, EventDescriptor, \
    StreamQueue, StreamQueueStrategy
from hopeit.app.context import EventContext, PostprocessHook, PreprocessHook
from hopeit.app.client import register_app_connections, stop_app_connections
from hopeit.dataobjects import DataObject, EventPayload
from hopeit.server.config import ServerConfig
from hopeit.server.events import EventHandler, get_event_settings, get_runtime_settings
from hopeit.streams import stream_auth_info, StreamEvent, StreamOSError, StreamManager
from hopeit.server.logger import engine_logger, extra_logger, combined
from hopeit.server.metrics import metrics, stream_metrics, StreamStats

__all__ = ['AppEngine',
           'Server']

logger = engine_logger()
extra = extra_logger()


class AppEngine:
    """
    Engine that handles a Hopeit Application
    """

    def __init__(self, *, app_config: AppConfig, plugins: List[AppConfig],
                 enabled_groups: List[str], streams_enabled: bool = True):
        """
        Creates an instance of the AppEngine

        :param app_config: AppConfig, Hopeit application configuration as specified in config module
        :param plugins: List of AppConfig, Hopeit application configurations for enabled plugins
        :enabled_groups: List of str, list of enabled event groups
        :streams_enabled: bool, for testing, set to False to disable automatic starting streams
        """
        self.app_config = app_config
        self.app_key = app_config.app_key()
        self.effective_events = self._config_effective_events(app_config, enabled_groups)
        self.plugins = plugins
        self.settings = get_runtime_settings(app_config, plugins)
        self.event_handler: Optional[EventHandler] = None
        self.streams_enabled = streams_enabled
        self.stream_manager: Optional[StreamManager] = None
        self._running: Dict[str, asyncio.Lock] = {
            event_name: asyncio.Lock()
            for event_name, event_info in self.effective_events.items()
            if event_info.type in (EventType.STREAM, EventType.SERVICE)
        }
        logger.init_app(app_config, plugins)

    async def start(self):
        """
        Starts handlers, services and pools for this application
        """
        self.event_handler = EventHandler(
            app_config=self.app_config, plugins=self.plugins,
            effective_events=self.effective_events,
            settings=self.settings)
        streams_present = any(
            True for _, event_info in self.effective_events.items()
            if (event_info.type == EventType.STREAM) or (event_info.write_stream is not None)
        )
        if streams_present and self.streams_enabled:
            mgr = StreamManager.create(self.app_config.server.streams)
            self.stream_manager = await mgr.connect()
        auth.init(self.app_key, self.app_config.server.auth)
        await register_app_connections(self.app_config)
        return self

    async def stop(self):
        """
        Stops and clean handlers
        """
        logger.info(__name__, f"Stopping app={self.app_key}...")
        for event_name, running in self._running.items():
            if running.locked():
                await self.stop_event(event_name)
        if self.stream_manager:
            await asyncio.sleep(self.app_config.engine.read_stream_timeout + 5)
            await self.stream_manager.close()
        await stop_app_connections(self.app_key)
        logger.info(__name__, f"Stopped app={self.app_key}")

    async def execute(self, *,
                      context: EventContext,
                      query_args: Optional[dict],
                      payload: Optional[EventPayload]) -> Optional[EventPayload]:
        """
        Executes a configured Event of type GET or POST using received payload as input,
        considering configured timeout.

        :param context: EventContext, info about app, event and tracking
        :param query_args: dict, containing query arguments to be passed to every step of event
        :param payload: EventPayload, payload to send to event handler
        :return: EventPayload, result from executing the event
        :raise: TimeoutException in case configured timeout is exceeded before getting the result
        """
        try:
            return await asyncio.wait_for(
                self._execute_event(context, query_args, payload),
                timeout=context.settings.response_timeout
            )
        except asyncio.TimeoutError as e:
            raise asyncio.TimeoutError(
                f"Response timeout exceeded seconds={context.settings.response_timeout}"
            ) from e

    async def _execute_event(self, context: EventContext,
                             query_args: Optional[dict],
                             payload: Optional[EventPayload],
                             queue: str = StreamQueue.AUTO) -> Optional[EventPayload]:
        """
        Process results of executing event specified in context for a given input payload.
        In case event yield multiple results (i.e. Spawn[...]) they are collected in batches
        of configured batch_size and written into output stream if configured.

        :return: result of executing the event. In case of multiple results yield from event,
        last item will be returned. If no items are yield, None will be returned.
        """
        assert self.event_handler is not None, "event_handler not created. Call `start()`."
        if self.streams_enabled and (context.event_info.write_stream is not None):
            assert self.stream_manager, "stream_manager not initialized. Call `start()`."
        event_info = self.effective_events[context.event_name]
        batch_size = context.settings.stream.batch_size
        batch = []
        result = None
        async for result in self.event_handler.handle_async_event(
                context=context, query_args=query_args, payload=payload):
            if result is not None:
                batch.append(result)
            if (len(batch) >= batch_size) and (event_info.write_stream is not None):
                await self._write_stream_batch(
                    batch=batch, context=context, event_info=event_info, queue=queue
                )
                batch.clear()
        if (len(batch) > 0) and (event_info.write_stream is not None):
            await self._write_stream_batch(
                batch=batch, context=context, event_info=event_info, queue=queue
            )
        return result

    async def _write_stream_batch(self, *, batch: List[EventPayload],
                                  context: EventContext,
                                  event_info: EventDescriptor,
                                  queue: str):
        await asyncio.gather(
            *[
                self._write_stream(
                    payload=item,
                    context=context,
                    event_info=event_info,
                    upstream_queue=queue
                ) for item in batch
            ]
        )

    async def _write_stream(self, *, payload: EventPayload,
                            context: EventContext,
                            event_info: EventDescriptor,
                            upstream_queue: str):
        """
        Publish payload in configured one or more queues for a given configured stream
        """
        assert self.stream_manager is not None, "stream_manager not created. Call `start()`."
        assert event_info.write_stream is not None, "write_stream name not configured"
        assert context.settings.stream.compression, "stream compression not configured"
        assert context.settings.stream.serialization, "stream serialization not configured"

        for configured_queue in event_info.write_stream.queues:
            stream_name = event_info.write_stream.name
            if (
                upstream_queue != StreamQueue.AUTO and
                configured_queue == StreamQueue.AUTO and
                event_info.write_stream.queue_strategy == StreamQueueStrategy.PROPAGATE
            ):
                stream_name += f".{upstream_queue}"
            elif configured_queue != StreamQueue.AUTO:
                stream_name += f".{configured_queue}"

            queue_name = (
                configured_queue
                if event_info.write_stream.queue_strategy == StreamQueueStrategy.DROP
                else upstream_queue
            )

            await self.stream_manager.write_stream(
                stream_name=stream_name,
                queue=queue_name,
                payload=StreamManager.as_data_event(payload),
                track_ids=context.track_ids,
                auth_info=context.auth_info,
                compression=context.settings.stream.compression,
                serialization=context.settings.stream.serialization,
                target_max_len=context.settings.stream.target_max_len
            )

    async def preprocess(self, *,
                         context: EventContext,
                         query_args: Optional[Dict[str, Any]],
                         payload: Optional[EventPayload],
                         request: PreprocessHook) -> Optional[EventPayload]:
        assert self.event_handler, "event_handler not created. Call `start()`."
        return await self.event_handler.preprocess(
            context=context, query_args=query_args, payload=payload, request=request)

    async def postprocess(self, *,
                          context: EventContext,
                          payload: Optional[EventPayload],
                          response: PostprocessHook) -> Optional[EventPayload]:
        assert self.event_handler, "event_handler not created. Call `start()`."
        return await self.event_handler.postprocess(context=context, payload=payload, response=response)

    async def _process_stream_event_with_timeout(
            self, stream_event: StreamEvent, stream_info: ReadStreamDescriptor,
            stream_name: str, queue: str,
            context: EventContext,
            stats: StreamStats,
            log_info: Dict[str, str]) -> Union[EventPayload, Exception]:
        """
        Invokes _process_stream_event with a configured timeout
        :return: result of _process_stream_event
        :raise: TimeoutError in case the event is not processed on the configured timeout
        """
        try:
            return await asyncio.wait_for(
                self._process_stream_event(stream_event=stream_event, stream_info=stream_info,
                                           stream_name=stream_name, queue=queue,
                                           context=context, stats=stats, log_info=log_info),
                timeout=context.settings.stream.timeout
            )
        except (asyncio.TimeoutError, asyncio.CancelledError):
            terr = asyncio.TimeoutError(
                f'Stream processing timeout exceeded seconds={context.settings.stream.timeout}'
            )
            logger.error(context, str(terr), extra=extra(
                prefix='stream.', **{**log_info, 'name': stream_name, 'queue': queue}
            ))
            return terr

    async def _read_stream_cycle(
            self,
            event_name: str,
            event_settings: EventSettings,
            stream_info: ReadStreamDescriptor,
            datatypes: Dict[str, type],
            offset: str,
            stats: StreamStats,
            log_info: Dict[str, str],
            test_mode: bool,
            last_err: Optional[StreamOSError]
    ) -> Tuple[Optional[Union[EventPayload, Exception]], Optional[EventContext], Optional[StreamOSError]]:
        """
        Single read_stream cycle used from read_stream while loop to allow wait and retry/recover on failures
        Will read from multiple queues if configured, always starting from the first queue and stopping
        when batch_size is reached
        """
        assert self.stream_manager is not None
        assert stream_info.consumer_group is not None

        last_res, last_context = None, None

        try:
            batch: List[Awaitable[Union[EventPayload, Exception]]] = []

            for queue in stream_info.queues:

                stream_name = stream_info.name
                if queue != StreamQueue.AUTO:
                    stream_name += f".{queue}"

                for stream_event in await self.stream_manager.read_stream(
                        stream_name=stream_name,
                        consumer_group=stream_info.consumer_group,
                        datatypes=datatypes,
                        track_headers=self.app_config.engine.track_headers,
                        offset=offset,
                        batch_size=event_settings.stream.batch_size,
                        timeout=self.app_config.engine.read_stream_timeout,
                        batch_interval=self.app_config.engine.read_stream_interval):

                    stats.ensure_start()

                    if isinstance(stream_event, Exception):
                        logger.error(__name__, stream_event)
                        stats.inc(error=True)
                    else:
                        context = EventContext(
                            app_config=self.app_config,
                            plugin_config=self.app_config,
                            event_name=event_name,
                            settings=event_settings,
                            track_ids=stream_event.track_ids,
                            auth_info=stream_auth_info(stream_event)
                        )
                        last_context = context
                        logger.start(context, extra=extra(
                            prefix='stream.', **{**log_info, 'name': stream_name, 'queue': queue}
                        ))
                        batch.append(self._process_stream_event_with_timeout(
                            stream_event=stream_event,
                            stream_info=stream_info,
                            stream_name=stream_name,
                            queue=stream_event.queue,
                            context=context,
                            stats=stats, log_info=log_info))

            if len(batch) != 0:
                for result in await asyncio.gather(*batch):
                    last_res = result
            if last_context:
                logger.stats(last_context, extra=extra(prefix='metrics.stream.', **stats.calc()))
            if test_mode:
                self._running[event_name].release()
            if last_err is not None:
                logger.warning(__name__, f"Recovered read stream for event={event_name}.")
                last_err = None

        except StreamOSError as e:
            retry_in = self.app_config.engine.read_stream_interval // 1000
            retry_in += random.randint(1, max(3, min(30, retry_in)))
            if type(e) != type(last_err):  # pylint: disable=unidiomatic-typecheck
                logger.error(__name__, e)
            logger.error(__name__, f"Cannot read stream for event={event_name}. Waiting seconds={int(retry_in)}...")
            last_err = e
            await asyncio.sleep(retry_in)  # TODO: Replace with circuit breaker

        return last_res, last_context, last_err

    async def read_stream(self, *,
                          event_name: str,
                          test_mode: bool = False) -> Optional[Union[EventPayload, Exception]]:
        """
        Listens to a stream specified by event of type STREAM, and executes
        the event handler for each received event in the stream.

        When invoked, stream will be read continuously consuming events that are
        arriving, processing it according to configured steps.
        To interrupt listening for events, call `stop_event(event_name)`

        :param event_name: str, an event name contained in app_config
        :param test_mode: bool, set to True to immediately stop and return results for testing
        :return: last result or exception, only intended to be used in test_mode
        """
        assert self.app_config.server is not None
        stats = StreamStats()
        log_info = {
            'app_key': self.app_key,
            'event_name': event_name
        }
        wait = self.app_config.server.streams.delay_auto_start_seconds
        if wait > 0:
            wait = int(wait/2) + random.randint(0, wait) - random.randint(0, int(wait/2))
            logger.info(__name__, f"Start reading stream: waiting seconds={wait}...",
                        extra=extra(prefix='stream.', **log_info))
            await asyncio.sleep(wait)
        logger.info(__name__, "Starting reading stream...", extra=extra(prefix='stream.', **log_info))
        try:
            assert self.event_handler, "event_handler not created. Call `start()`."
            assert self.stream_manager, "No active stream manager. Call `start()`"
            assert not self._running[event_name].locked(), "Event already running. Call `stop_event(...)`"

            event_config = self.effective_events[event_name]
            stream_info = event_config.read_stream
            assert stream_info, f"No read_stream section in config for event={event_name}"
            event_settings = get_event_settings(self.settings, event_name)
            for queue in stream_info.queues:
                await self.stream_manager.ensure_consumer_group(
                    stream_name=(
                        f"{stream_info.name}.{queue}"
                        if queue != StreamQueue.AUTO
                        else stream_info.name
                    ),
                    consumer_group=stream_info.consumer_group
                )

            datatypes = self._find_stream_datatype_handlers(event_name, event_config)
            log_info['name'] = stream_info.name
            log_info['consumer_group'] = stream_info.consumer_group
            assert not self._running[event_name].locked(), f"Event already running {event_name}"
            await self._running[event_name].acquire()
            logger.info(__name__, "Consuming stream...", extra=extra(prefix='stream.', **log_info))
            offset = '>'
            last_res, last_context, last_err = None, None, None
            while self._running[event_name].locked():
                last_res, last_context, last_err = await self._read_stream_cycle(
                    event_name, event_settings, stream_info, datatypes, offset, stats,
                    log_info, test_mode, last_err
                )
            logger.info(__name__, 'Stopped read_stream.', extra=extra(prefix='stream.', **log_info))
            if last_context is None:
                logger.warning(__name__, f"No stream events consumed in {event_name}")
            return last_res
        except (AssertionError, NotImplementedError) as e:
            logger.error(__name__, e)
            logger.error(__name__, f"Unexpectedly stopped read stream for event={event_name}")
            return e

    async def _process_stream_event(self, *, stream_event: StreamEvent,
                                    stream_info: ReadStreamDescriptor,
                                    stream_name: str,
                                    queue: str,
                                    context: EventContext,
                                    stats: StreamStats,
                                    log_info: Dict[str, str]) -> Optional[Union[EventPayload, Exception]]:
        """
        Process a single stream event, execute events, ack if not failed, log error if fail

        :return: results of executing the event, or Exception if errors during processing
        """
        try:
            assert self.event_handler
            assert self.stream_manager

            result = await self._execute_event(
                context=context, query_args=None,
                payload=stream_event.payload, queue=stream_event.queue
            )
            await self.stream_manager.ack_read_stream(
                stream_name=stream_name,
                consumer_group=stream_info.consumer_group,
                stream_event=stream_event
            )
            logger.done(context, extra=combined(
                extra(prefix='stream.', **{**log_info, 'name': stream_name, 'queue': queue}),
                metrics(context),
                stream_metrics(context)
            ))
            stats.inc()
            return result
        except CancelledError as e:
            extra_info = {**log_info, 'name': stream_name, 'queue': queue}
            logger.error(context, 'Cancelled', extra=extra(prefix='stream.', **extra_info))
            logger.failed(context, extra=extra(prefix='stream.', **extra_info))
            stats.inc(error=True)
            return e
        except Exception as e:  # pylint: disable=broad-except
            extra_info = {**log_info, 'name': stream_name, 'queue': queue}
            logger.error(context, e, extra=extra(prefix='stream.', **extra_info))
            logger.failed(context, extra=extra(prefix='stream.', **extra_info))
            stats.inc(error=True)
            return e

    def _service_event_context(self, event_name: str, event_settings: EventSettings,
                               previous_context: Optional[EventContext] = None):
        if previous_context is None:
            track_ids = {
                'track.request_id': str(uuid.uuid4()),
                'track.request_ts': datetime.now(tz=timezone.utc).isoformat()
            }
        else:
            track_ids = previous_context.track_ids
        return EventContext(
            app_config=self.app_config,
            plugin_config=self.app_config,
            event_name=event_name,
            settings=event_settings,
            track_ids={
                **track_ids,
                'track.operation_id': str(uuid.uuid4()),
            },
            auth_info={}
        )

    async def service_loop(self, *,
                           event_name: str,
                           test_mode: bool = False) -> Optional[Union[EventPayload, Exception]]:
        """
        Service loop, executes `__service__` handler in event and execute
        event steps for each yielded payload.

        :param event_name: str, an event name contained in app_config
        :param test_mode: bool, set to True to immediately stop and return results for testing
        :return: last result or exception, only intended to be used in test_mode
        """
        assert self.app_config.server is not None
        log_info = {
            'app_key': self.app_key,
            'event_name': event_name
        }
        assert not self._running[event_name].locked(), f"Cannot start service, event already running {event_name}"
        await self._running[event_name].acquire()
        wait = self.app_config.server.streams.delay_auto_start_seconds
        if wait > 0:
            wait = int(wait/2) + random.randint(0, wait) - random.randint(0, int(wait/2))
            logger.info(__name__, f"Start service: waiting seconds={wait}...",
                        extra=extra(prefix='service.', **log_info))
            await asyncio.sleep(wait)
        logger.info(__name__, "Starting service...", extra=extra(prefix='service.', **log_info))
        event_config = self.effective_events[event_name]
        impl = find_event_handler(app_config=self.app_config, event_name=event_name, event_info=event_config)
        service_handler = getattr(impl, '__service__')
        assert service_handler is not None, \
            f"{event_name} must implement method `__service__(context) -> Spawn[...]` to run as a service"
        event_settings = get_event_settings(self.settings, event_name)
        context = self._service_event_context(event_name=event_name, event_settings=event_settings)
        last_result = None
        if self._running[event_name].locked():
            async for payload in service_handler(context):
                try:
                    context = self._service_event_context(
                        event_name=event_name, event_settings=event_settings, previous_context=context
                    )
                    logger.start(context, extra=extra(prefix='service.', **log_info))
                    last_result = await self.execute(
                        context=context, query_args=None, payload=payload
                    )
                    logger.done(context, extra=extra(prefix='service.', **log_info))
                    if not self._running[event_name].locked():
                        logger.info(__name__, "Stopped service.", extra=extra(prefix='service.', **log_info))
                        break
                except CancelledError as e:
                    logger.error(context, 'Cancelled', extra=extra(prefix='service.', **log_info))
                    logger.failed(context, extra=extra(prefix='service.', **log_info))
                    last_result = e
                except Exception as e:  # pylint: disable=broad-except
                    logger.error(context, e, extra=extra(prefix='service.', **log_info))
                    logger.failed(context, extra=extra(prefix='service.', **log_info))
                    last_result = e
                if test_mode:
                    self._running[event_name].release()
                    return last_result
        else:
            logger.info(__name__, "Stopped service.", extra=extra(prefix='service.', **log_info))
        logger.info(__name__, "Finished service.", extra=extra(prefix='service.', **log_info))
        return last_result

    def is_running(self, event_name) -> bool:
        return self._running[event_name].locked()

    async def stop_event(self, event_name: str):
        """
        Sets running state to stopped for a continuous-running event.
        This acts as signling for stop for STREAM events.

        :param event_name: name of the event to signal stop
        """
        if self._running[event_name].locked():
            self._running[event_name].release()
        else:
            raise RuntimeError(f"Cannot stop non running event: {event_name}.")

    @staticmethod
    def _config_effective_events(app_config: AppConfig, enabled_groups: List[str]) -> Dict[str, EventDescriptor]:
        """
        Return effective events computed from user app config.

        Effective events could be result of splitting a single event in stages,
        using the "SHUFFLE" keyword, that will internaally generate 2 events.
        Or for STREAMS that implements the `__service__` method, both a STREAM
        and a SERVICE event will be generated.
        Only events with groups defined in `enabled_groups` list will be returned.
        If `enabled_groups` is empty, all events wil be considered.
        """
        effective_events: Dict[str, EventDescriptor] = {}
        assert app_config.server
        for event_name, event_info in app_config.events.items():
            if (
                len(enabled_groups) == 0
                or event_info.group == EventDescriptor.DEFAULT_GROUP
                or event_info.group in enabled_groups
            ):
                impl = find_event_handler(app_config=app_config, event_name=event_name, event_info=event_info)
                # Add events resultant of splitting steps on SHUFFLE (stages)
                splits = split_event_stages(app_config.app, event_name, event_info, impl)
                effective_events.update(**splits)
                # Add associated SERVICE events to streams
                if event_info.type == EventType.STREAM and hasattr(impl, "__service__"):
                    effective_events[f"{event_name}$__service__"] = EventDescriptor(
                        type=EventType.SERVICE,
                        connections=event_info.connections,
                        impl=event_info.impl,
                    )
        return effective_events

    def _find_stream_datatype_handlers(self, event_name: str, event_info: EventDescriptor) -> Dict[str, type]:
        """
        Computes a dictionary of `{datatype name: datatype class}` that event steps
        can be handle when consuming from an stream.
        """
        base_event, _ = event_and_step(event_name)
        impl = find_event_handler(app_config=self.app_config, event_name=base_event, event_info=event_info)
        all_steps = extract_module_steps(impl)
        steps = effective_steps(event_name, all_steps)
        datatypes = {}
        for _, _, step in steps:
            _, datatype, _, _ = step
            if hasattr(datatype, '__stream_event__'):
                datatypes[f"{datatype.__module__}.{datatype.__qualname__}"] = datatype
            elif datatype is DataObject:
                for type_name in event_info.dataobjects:
                    datatype = find_datobject_type(type_name)
                    datatypes[f"{datatype.__module__}.{datatype.__qualname__}"] = datatype
        if len(datatypes) == 0:
            raise NotImplementedError(f"No data types found to read from stream in event={event_name}. "
                                      "Dataclasses must be decorated with `@dataobject` to be used in streams")
        return datatypes


class Server:
    """
    Server engine.
    Call `start()` to create an instance of the engine
    Then start individual apps by calling `start_app(...)`
    End apps and stop engine with `stop()`
    """
    def __init__(self):
        self.app_engines = {}

    async def start(self, *, config: ServerConfig):
        """
        Starts a engine instance
        """
        global logger
        logger = engine_logger().init_server(config)
        logger.info(__name__, 'Starting engine...')
        return self

    async def stop(self):
        """
        Stops every active app in the engine instance.
        """
        logger.info(__name__, 'Stopping engine...')
        for _, app_engine in self.app_engines.items():
            await app_engine.stop()
        logger.info(__name__, 'Engine stopped.')

    async def start_app(self, app_config: AppConfig, enabled_groups: List[str]):
        """
        Starts and register a Hopeit App into this engine instance

        :param app_config: AppConfig, app configuration as specified in config module
        """
        logger.info(__name__, f"Starting app={app_config.app_key()}...")
        plugins = [
            self.app_engine(app_key=plugin.app_key()).app_config
            for plugin in app_config.plugins
        ]
        app_engine = await AppEngine(app_config=app_config,
                                     plugins=plugins,
                                     enabled_groups=enabled_groups).start()
        self.app_engines[app_config.app_key()] = app_engine
        return app_engine

    def app_engine(self, *, app_key: str) -> AppEngine:
        return self.app_engines[app_key]



"""
Special handled exceptions to return specific response status.
All other exceptions will return 500 (Internal Server Error)
"""
__all__ = ['BadRequest',
           'Unauthorized']

from hopeit.server import errors


class BadRequest(BaseException):
    """
    Exception that will return BadRequest (400) response in endpoint
    """
    ErrorInfo = errors.ErrorInfo  # pylint: disable=invalid-name


class Unauthorized(BaseException):
    """
    Exception that will return Unauthorized (401) response in endpoint
    """
    ErrorInfo = errors.ErrorInfo  # pylint: disable=invalid-name



"""
Events handling
"""
from typing import AsyncGenerator, Type, TypeVar

from hopeit.dataobjects import EventPayloadType
from hopeit.server.collector import Collector
from hopeit.server.steps import SHUFFLE, CollectorStepsDescriptor

__all__ = ['Spawn', 'SHUFFLE', 'collector_step', 'Collector']

T = TypeVar('T')
Spawn = AsyncGenerator[T, None]


def collector_step(*, payload: Type[EventPayloadType]) -> CollectorStepsDescriptor:
    """
    Specifies a set of steps that would be run using `AsyncCollector` implementation. AsyncCollector ensures
    all steps are run as concurrently as possible using `asyncio.gather(...)` mechanism.
    Each step has to implement the signature: `step(collector: Collector, context: EventContext)`
    and the data returned by the step will be put in the collector.
    Each step should await for data in the collector using `data = await collector['step_name']`.
    This way collector guarantees that dependant steps will be executed before steps
    using generated data in the collector.

    :param payload: EventPayloadType, collector contains a special bucket for "payload" information
        that will be submitted on evey call and can be accessed using `payload = await collector['payload']`,
        so Type of payload needs to be informed to create the appropriate handler: collector(payload=DataType)

    Example:

    Use inline in __steps__ definition of event implementation,
    to run step1, step2 and step3 concurrently using a Collector, and then
    seq_step4, and seq_step5 sequentially::

        __steps__ = [
            collector_step(payload=InputType).gather('step1', 'step2', 'step3),
            'seq_step4', 'seq_step5')
        ]

        async def step1(collector: Collector, context: EventContext) -> StepData:
            payload = await collector['payload']
            # do something with payload, this can run concurrently with step2
            return StepData("some data")

        async def step2(collector: Collector, context: EventContext) -> StepData:
            payload = await collector['payload']
            # do something with payload, this can run concurrently with step1
            return StepData("some data")

        async def step3(collector: Collector, context: EventContext) -> StepData:
            step1data = await collector['step1']
            step2data = await collector['step2']
            # do something with step1data and step2data, step1 and step2 must be completed at this point
            return StepData("some combined data")

        async def seq_step4(collector: Collector, context: EventContext) -> SeqStepData:
            step3data = await collector['step3']
            # do something with step3data from the collector
            return SeqStepData("some data")

        async def seq_step5(payload: SeqStepData, context: EventContext) -> SeqStepData:
            # do something with payload (no need to use collector anymore)
            return SeqStepData("some data")

    Notice that collector steps can be combined with regular sequential steps. Each collector will create
    internally a single sequential step to be handled by the engine while the collector itself will
    handle execution for the steps specified in .gather(...) definition. After collector is executed
    the next sequential after the collector step will receive the `collector: Collector` as first argument.

    CAUTION: AsyncCollector makes not guarantees whether your code will block indefinitely (i,e.
    if you do `await collector['step1']` from step2func but step1func does `await collector['step2']`. This
    will block your application, so the only way the engine will prevent this to lock your server is
    using timeouts. In case of a dead-lock, event will fail to process and concurrent functions
    will be canceled when reaching a timeout, but no checking is done on whether a deadlock is happening.
    Please check the sequence your code is accessing/awaiting results from the collector to avoid cycles.
    """
    return CollectorStepsDescriptor(payload)



"""
Simple encryption tools for testing development
"""
from pathlib import Path
from typing import Union

from cryptography.fernet import Fernet

__all__ = [
    'setup_encryption',
    'encrypt',
    'decrypt'
]

key_file: Path = Path.home() / '.hopeit' / 'test.key'


def setup_encryption(path_to_key_file: Union[str, Path]):
    global key_file
    if isinstance(path_to_key_file, str):
        path_to_key_file = Path(path_to_key_file)
    key_file = path_to_key_file


def encrypt(text):
    with open(key_file, encoding="utf-8") as f:
        return Fernet(f.read()).encrypt(text.encode()).decode()


def decrypt(text):
    with open(key_file, encoding="utf-8") as f:
        return Fernet(f.read()).decrypt(text.encode()).decode()



"""
Open API spec creation and server helpers
"""
import json
import re
from copy import deepcopy
from functools import partial
from pathlib import Path
from typing import Dict, List, Tuple, Type, Optional, Callable, Awaitable, Union
from datetime import date, datetime

from aiohttp import web
from aiohttp_swagger3 import RapiDocUiSettings  # type: ignore
from aiohttp_swagger3.swagger import Swagger  # type: ignore
from aiohttp_swagger3.exceptions import ValidatorError  # type: ignore
from aiohttp_swagger3 import validators  # type: ignore
from aiohttp_swagger3.validators import MISSING, _MissingType  # type: ignore
from aiohttp_swagger3.swagger_route import SwaggerRoute  # type: ignore
from stringcase import titlecase  # type: ignore
import typing_inspect as typing  # type: ignore
from dataclasses_jsonschema import SchemaType

from hopeit.dataobjects import BinaryAttachment, BinaryDownload  # type: ignore
from hopeit.app.config import AppConfig, AppDescriptor, EventDescriptor, EventPlugMode, EventType
from hopeit.server.config import ServerConfig, AuthType
from hopeit.server.errors import ErrorInfo
from hopeit.server.imports import find_event_handler
from hopeit.server.logger import engine_logger
from hopeit.server.names import route_name
from hopeit.server.steps import extract_module_steps, extract_postprocess_handler, extract_preprocess_handler, \
    StepInfo


__all__ = ['init_empty_spec',
           'load_api_file',
           'save_api_file',
           'setup',
           'clear',
           'app_route_name',
           'register_server_config',
           'register_apps',
           'enable_swagger',
           'diff_specs']

logger = engine_logger()

swagger: Optional[Swagger] = None
spec: Optional[dict] = None
static_spec: Optional[dict] = None
runtime_schemas = {}
_options = {
    'generate_mode': False
}

OPEN_API_VERSION = '3.0.3'

METHOD_MAPPING = {
    EventType.GET: 'get',
    EventType.POST: 'post',
    EventType.MULTIPART: 'post'
}


class APIError(Exception):
    """
    Error thrown when API incompatibilities are detected
    """


def setup(**kwargs):
    """
    Setup additional options for api module. Supported options are:

    :param generate_mode: bool, default False: creates empty path placholders for modules not defining __api__
        specification
    """
    _options.update(**kwargs)


def clear():
    """
    Clears api configuration stored in memory. This disables api module.
    """
    global spec, static_spec, swagger, runtime_schemas, _options
    spec = None
    static_spec = None
    swagger = None
    runtime_schemas = {}
    _options = {
        'generate_mode': False
    }


def init_empty_spec(api_version: str, title: str, description: str):
    """
    Initializes internal spec and static_spec dictionaries with minimal Open API requirements:
    openapi, info sections and empty paths. This method can be used to create new API specs.
    :param api_version: info.version
    :param title: info.title
    :param description: info.description
    """
    global spec, static_spec
    logger.info(__name__, "Creating Open API spec...")
    spec = {
        "openapi": OPEN_API_VERSION,
        "info": {
            "version": api_version,
            "title": title,
            "description": description
        },
        "paths": {}
    }
    logger.info(__name__, f"API: openapi={spec['openapi']}, API version={spec['info']['version']}")
    static_spec = deepcopy(spec)


def load_api_file(path: Union[str, Path]):
    """
    Loads OpenAPI spec from a json file. Spec is loaded into the module.
    @param path: path to json file
    """
    global spec, static_spec
    logger.info(__name__, f"Loading api spec from api_file={path}...")
    with open(path, 'r', encoding="utf-8") as f:
        spec = json.loads(f.read())
        assert spec is not None
        logger.info(__name__, f"API: openapi={spec['openapi']}, API version={spec['info']['version']}")
        static_spec = deepcopy(spec)


def save_api_file(path: Union[str, Path], api_version: str):
    """
    Saves module Open API spec to json file.
    :param path: path to json file
    :param api_version: new api_version, in case changes between previously loaded api file and api calculated at
        runtime, a new api_version needs to be specified to allow saving the file.
    """
    assert spec is not None
    assert static_spec is not None
    if diff_specs() and static_spec['info']['version'] == api_version:
        err = APIError("Cannot save api file. Need to increment version number. Differences found.")
        logger.error(__name__, err)
        raise err
    logger.info(__name__, f"Set API version={api_version}...")
    spec['info']['version'] = api_version
    logger.info(__name__, f"Saving api spec to api_file={path}...")
    with open(path, 'w', encoding="utf-8") as f:
        f.write(json.dumps(spec, indent=2))
        f.flush()


def register_server_config(server_config: ServerConfig):
    """
    Register API definitions from server configuration. This consists of allowed and default authentication methods.
    """
    if spec is not None:
        if 'components' not in spec:
            spec['components'] = {'schemas': {}}
        _update_auth_methods()
        _update_server_default_auth_methods(server_config)


def register_apps(apps_config: List[AppConfig]):
    """
    Register api definition for a list of apps that conform to a single API specification.

    @param apps_config: list of AppConfig objects to be introspected
    """
    if spec is not None:
        logger.info(__name__, "Registering apps...")
        apps_config_by_key = {config.app.app_key(): config for config in apps_config}
        for config in apps_config:
            logger.info(__name__, f"Updating API spec for app={config.app_key()}...")
            _register_api_spec(config)
            for plugin in config.plugins:
                logger.info(__name__, f"Updating API spec for app={config.app_key()}, plugin={plugin.app_key()}...")
                plugin_config = apps_config_by_key[plugin.app_key()]
                _register_api_spec(config, plugin_config)
        _cleanup_api_schemas()
        _cleanup_global_auth()


def _register_api_spec(app_config: AppConfig, plugin: Optional[AppConfig] = None):
    if spec is not None:
        if 'components' not in spec:
            spec['components'] = {'schemas': {}}
        _update_predefined_schemas()
        _update_api_schemas(app_config)
        _update_api_paths(app_config, plugin)


def diff_specs() -> bool:
    """
    Detects differences between loaded API specification and spec calculated from server and apps.

    :return: True if differences are found, False if loaded spec matches runtime.
    """
    return static_spec != spec


async def _passthru_handler(request: web.Request) -> Tuple[web.Request, bool]:
    return request, True


class CustomizedObjectValidator(validators.Object):  # pragma: no cover
    """
    Replacements of Object Validator provided by aiohttp3_swagger
    to handle multipart form requests
    """
    def validate(self, raw_value: Union[None, Dict, _MissingType],
                 raw: bool) -> Union[None, Dict, _MissingType]:
        # FIXED: is_missing = isinstance(raw_value, _MissingType)
        is_missing = (
            isinstance(raw_value, _MissingType)
            or ((raw_value is not None) and (not isinstance(raw_value, dict)))
        )
        # ORIGINAL CODE: https://github.com/hh-h/aiohttp-swagger3/blob/master/aiohttp_swagger3/validators.py
        # FIXED END
        if not is_missing and self.readOnly:
            raise ValidatorError("property is read-only")
        if raw_value is None:
            if self.nullable:
                return None
            raise ValidatorError("value should be type of dict")
        if not isinstance(raw_value, dict):
            if is_missing:
                return raw_value
            raise ValidatorError("value should be type of dict")
        value = {}
        errors: Dict = {}
        for name in self.required:
            if name not in raw_value:
                errors[name] = "required property"
        if errors:
            raise ValidatorError(errors)

        for name, validator in self.properties.items():
            prop = raw_value.get(name, MISSING)
            try:
                val = validator.validate(prop, raw)
                if val != MISSING:
                    value[name] = val
            except ValidatorError as e:
                errors[name] = e.error
        if errors:
            raise ValidatorError(errors)

        if isinstance(self.additionalProperties, bool):
            if not self.additionalProperties:
                additional_properties = raw_value.keys() - value.keys()
                if additional_properties:
                    raise ValidatorError({k: "additional property not allowed" for k in additional_properties})
            else:
                for key in raw_value.keys() - value.keys():
                    value[key] = raw_value[key]
        else:
            for name in raw_value.keys() - value.keys():
                validator = self.additionalProperties
                value[name] = validator.validate(raw_value[name], raw)
        if self.minProperties is not None and len(value) < self.minProperties:
            raise ValidatorError(f"number or properties must be more than {self.minProperties}")
        if self.maxProperties is not None and len(value) > self.maxProperties:
            raise ValidatorError(f"number or properties must be less than {self.maxProperties}")
        return value


setattr(validators.Object, "validate", CustomizedObjectValidator.validate)


def enable_swagger(server_config: ServerConfig, app: web.Application):
    """
    Enables Open API (a.k.a Swagger) on this server. This consists of:
        * All endpoints within API specification are to be handled by a Open API handler that will validate requests
        * If specified in server_config.api_docs_path, API docs site will be available at the given route.
            i.e. http://server-address:8020/api/docs
    :param server_config: server configuration
    :param app: aiohttp web Application to host routes and docs
    """
    global swagger, static_spec
    if spec is None:
        logger.warning(__name__, "No api-file loaded. OpenAPI docs and validation disabled.")
        return
    if diff_specs():
        err = APIError("Cannot enable OpenAPI. Differences found between api-file and running apps. "
                       "Run `hopeit openapi diff` to check and `hopeit openapi update` to generate spec file")
        logger.error(__name__, err)
        raise err
    static_spec = None
    logger.info(__name__, "Enabling OpenAPI endpoints...")
    app["AIOHTTP_SWAGGER3_SWAGGER_SPECIFICATION"] = spec
    api_docs_ui = None
    if server_config.api.docs_path:
        api_docs_ui = RapiDocUiSettings(
            path=server_config.api.docs_path,
            heading_text=spec['info']['title'],
            theme='dark',
            render_style='read',
            layout='column',
            schema_style='tree',
            allow_spec_url_load=False,
            allow_spec_file_load=False,
            allow_server_selection=False,
            show_header=False
        )
        logger.info(__name__, f"OpenAPI documentation available in {server_config.api.docs_path}")
    else:
        logger.warning(
            __name__, "OpenAPI documentation path not specified in server config. API docs endpoint disabled.")

    swagger = Swagger(
        app,
        validate=True,
        spec=spec,
        request_key="data",
        rapidoc_ui_settings=api_docs_ui,
        redoc_ui_settings=None,
        swagger_ui_settings=None
    )
    swagger.register_media_type_handler("multipart/form-data", _passthru_handler)
    logger.info(__name__, "OpenAPI validations enabled.")


def add_route(method: str,
              path: str,
              handler: Callable[..., Awaitable[web.StreamResponse]]) -> Callable[..., Awaitable[web.StreamResponse]]:
    """
    Register a route handler. In case the path is associated with a path in Open API running spec,
    handler is to be wrapped by an Open API handler, if not, handler will be returned with no changes
    and a WARNING is logged.

    :param method: str, valid Open API method (i.e. GET, POST)
    :param path: str, route
    :param handler: function to be used as handler
    """
    if spec is None:
        return handler
    assert swagger is not None, "API module not initialized. Call `api.enable_swagger(...)`"
    method_lower = method.lower()
    if method_lower in spec["paths"].get(path, {}):
        route = SwaggerRoute(method_lower, path, handler, swagger=swagger)
        api_handler = partial(swagger._handle_swagger_call, route)  # pylint: disable=protected-access
        return api_handler
    logger.warning(__name__, f"No API Spec defined for path={path}")
    return handler


def app_route_name(app: AppDescriptor, *, event_name: str,
                   plugin: Optional[AppDescriptor] = None,
                   prefix: str = 'api', override_route_name: Optional[str] = None) -> str:
    """
    Returns the full route name for a given app event

    :param app: AppDescriptor, as defined in AppConfig
    :param event_name: event name as defined in AppConfig
    :param plugin: optional plugin if the event comes from a plugin and EventPlugMode=='OnApp'
    :param prefix: route prefix, defaults to 'api'
    :param override_route_name: Optional[str], provided route to be used instead app and event name,
        if starts with '/', prefix will be ignored, otherwised appended to prefix
    :return: str, full route name. i.e.:
        /api/app-name/1x0/event-name or /api/app-name/1x0/plugin-name/1x0/event-name
    """
    components = [
        prefix, app.name, app.version,
        *([plugin.name, plugin.version] if plugin else []),
        *event_name.split('.')
    ] if override_route_name is None else [
        override_route_name[1:]
    ] if override_route_name[0] == '/' else [
        prefix, override_route_name
    ]
    return route_name(*components)


def _schema_name(datatype: type) -> str:
    return f"#/components/schemas/{datatype.__name__}"


def datatype_schema(event_name: str, datatype: Type) -> dict:
    origin = typing.get_origin(datatype)
    if origin is None:
        origin = datatype
    type_mapper = TYPE_MAPPERS.get(origin)
    if type_mapper is None:
        return {
            "$ref": _schema_name(datatype)
        }
    return type_mapper(event_name, datatype)  # type: ignore


def _update_auth_methods():
    """
    Generate default securitySchemes section
    """
    security_schemas = spec['components'].get('securitySchemes', {})
    security_schemas.update({
        'auth.basic': {
            'type': 'http',
            'scheme': 'basic'
        },
        'auth.bearer': {
            'type': 'http',
            'scheme': 'bearer'
        }
    })
    spec['components']['securitySchemes'] = security_schemas


def _update_auth_refresh_method(app_key: str):
    """
    Generate securitySchemes entries for REFRESH token cookie for each app
    """
    assert spec is not None
    security_schemas = spec['components'].get('securitySchemes', {})
    security_schemas.update({
        f"{app_key}.refresh": {
            'type': 'apiKey',
            'in': 'cookie',
            'name': f"{app_key}.refresh"
        }
    })
    spec['components']['securitySchemes'] = security_schemas


def _update_server_default_auth_methods(server_config: ServerConfig):
    """
    Generate security section based on server default_auth_methods
    """
    assert spec is not None
    security = spec.get('security', [])
    methods = {method for entry in security for method in entry.keys()}
    for auth_method in server_config.auth.default_auth_methods:
        auth_str = f"auth.{auth_method.value.lower()}"
        if auth_str != 'auth.unsecured' and auth_str not in methods:
            security.append({auth_str: []})
    spec['security'] = security


def _update_api_schemas(app_config: AppConfig):
    """
    Generate schemas for @dataobject annotated dataclasses discovered in event implementation modules
    """
    assert spec is not None
    schemas = spec['components'].get('schemas', {})
    for event_name, event_info in app_config.events.items():
        event_schemas = _generate_schemas(app_config, event_name, event_info)
        for name, event_schema in event_schemas.items():
            if name in runtime_schemas:
                if not event_schema == schemas.get(name):
                    logger.warning(__name__,
                                   f"Schema ignored: same schema name has non-compatible implementations: "
                                   f"event={event_name} schema={name}")
            else:
                schemas[name] = event_schema
                runtime_schemas[name] = event_schema

    spec['components']['schemas'] = schemas


def _update_predefined_schemas():
    """
    Generate schemas for predefined classes
    """
    assert spec is not None
    spec['components']['schemas'].update(
        ErrorInfo.json_schema(schema_type=SchemaType.V3, embeddable=True)
    )


def _cleanup_api_schemas():
    """
    Remove schemas from spec, if they are not used in paths
    """
    assert spec is not None
    modified = True
    while modified:
        clean = {}
        spec_str = json.dumps(spec)
        schemas = spec['components'].get('schemas', {})
        for name, schema in schemas.items():
            if spec_str.find(f"#/components/schemas/{name}") >= 0:
                clean[name] = schema
        modified = len(schemas) > len(clean)
        spec['components']['schemas'] = clean


def _cleanup_global_auth():
    """
    Remove global security requirements as they are propagated path by path.
    """
    assert spec is not None
    spec['security'] = []


def _update_api_paths(app_config: AppConfig, plugin: Optional[AppConfig] = None):
    """
    Populates paths section of spec based on __api__ specified in implemented events
    """
    assert spec is not None
    events = {
        k: v for k, v in app_config.events.items() if v.plug_mode == EventPlugMode.STANDALONE
    } if plugin is None else {
        k: v for k, v in plugin.events.items() if v.plug_mode == EventPlugMode.ON_APP
    }
    plugin_app = None if plugin is None else plugin.app
    paths = spec.get('paths', {})
    for event_name, event_info in events.items():
        route = app_route_name(app_config.app, event_name=event_name, plugin=plugin_app,
                               override_route_name=event_info.route)
        method = METHOD_MAPPING.get(event_info.type)
        if method is None:
            continue
        event_api_spec = _extract_event_api_spec(
            app_config if plugin is None else plugin, event_name, event_info
        )
        if event_api_spec is None:
            event_api_spec = paths.get(route, {}).get(method)
        if event_api_spec is None and _options.get('generate_mode'):
            event_api_spec = {"description": f"<<<{event_name}>>>", "parameters": [], "responses": {}}
        if event_api_spec is not None:
            event_api_spec['tags'] = [app_config.app_key()]
            _set_optional_fixed_headers(event_api_spec)
            _set_track_headers(event_api_spec, app_config)
            _set_path_security(event_api_spec, app_config, event_info)
            route_path = paths.get(route, {})
            route_path[method] = event_api_spec
            paths[route] = route_path
    spec['paths'] = paths


def _set_optional_fixed_headers(event_api_spec: dict):
    """
    Set arguments for request-id and request-ts track headers on every path entry
    """
    if not any(param['name'] == 'X-Track-Request-Id' for param in event_api_spec['parameters']):
        event_api_spec['parameters'].append({
            "name": "X-Track-Request-Id",
            "in": "header",
            "required": False,
            "description": "Track information: Request-Id",
            "schema": {
                "type": "string"
            }
        })
    if not any(param['name'] == 'X-Track-Request-Ts' for param in event_api_spec['parameters']):
        event_api_spec['parameters'].append({
            "name": "X-Track-Request-Ts",
            "in": "header",
            "required": False,
            "description": "Track information: Request-Ts",
            "schema": {
                "type": "string",
                "format": "date-time"
            }
        })


def _set_track_headers(event_api_spec: dict, app_config: AppConfig):
    """
    Set arguments for track headers specified in app_config for every path
    """
    current_params = {entry['name'] for entry in event_api_spec['parameters']}
    for track_header in app_config.engine.track_headers:
        header_name = f"X-{re.sub(' ', '-', titlecase(track_header))}"
        if header_name not in current_params:
            event_api_spec['parameters'].append({
                "name": header_name,
                "in": "header",
                "required": True,
                "description": f"Track information: {track_header}",
                "schema": {
                    "type": "string",
                    "default": track_header.replace('track', 'test')
                }
            })


def _set_path_security(event_api_spec: dict, app_config: AppConfig, event_info: EventDescriptor):
    """
    Setup security schemes allowed for each path
    """
    assert spec is not None
    security: list = []
    for auth in event_info.auth:
        if auth == AuthType.REFRESH:
            _update_auth_refresh_method(app_config.app_key())
            auth_str = f"{app_config.app_key()}.refresh"
            security.append({auth_str: []})
        elif auth != AuthType.UNSECURED:
            auth_str = f"auth.{auth.value.lower()}"
            security.append({auth_str: []})
    if len(security) == 0 and AuthType.UNSECURED not in event_info.auth:
        security = spec['security']
    if len(security) > 0:
        event_api_spec['security'] = security


def _extract_event_api_spec(app_config: AppConfig, event_name: str, event_info: EventDescriptor) -> Optional[dict]:
    """
    Extract __api__ definition from event implementation
    """
    module = find_event_handler(app_config=app_config, event_name=event_name, event_info=event_info)
    if hasattr(module, '__api__'):
        method_spec = getattr(module, '__api__')
        if isinstance(method_spec, dict):
            return method_spec
        return method_spec(module, app_config, event_name, None)
    return None


def _generate_schemas(app_config: AppConfig, event_name: str, event_info: EventDescriptor) -> dict:
    """
    Generate all schemas for a given event, based on steps signatures
    """
    module = find_event_handler(app_config=app_config, event_name=event_name, event_info=event_info)
    steps = extract_module_steps(module)
    schemas: dict = {}
    for _, step_info in steps:
        _update_step_schemas(schemas, step_info)
    step_info = extract_postprocess_handler(module)
    _update_step_schemas(schemas, step_info)
    step_info = extract_preprocess_handler(module)
    _update_step_schemas(schemas, step_info)
    return schemas


def _update_step_schemas(schemas: dict, step_info: Optional[StepInfo]):
    if step_info is not None:
        _, input_type, ret_type, _ = step_info
        datatypes = _explode_datatypes([input_type, ret_type])
        for datatype in datatypes:
            if datatype is not None and hasattr(datatype, '__data_object__'):
                if datatype.__data_object__['schema']:
                    schemas.update(datatype.json_schema(schema_type=SchemaType.V3, embeddable=True))


def _explode_datatypes(datatypes: List[Type]) -> List[Type]:
    result = []
    for datatype in datatypes:
        if datatype is not None:
            if hasattr(datatype, '__args__'):
                for arg in getattr(datatype, '__args__'):
                    result.extend(_explode_datatypes([arg]))
            else:
                result.append(datatype)
    return result


def _array_schema(event_name: str, datatype: type):
    args = typing.get_args(datatype)
    return {
        "type": "array",
        "items": {
            "$ref": _schema_name(args[0])
        }
    }


def _binary_download_schema(event_name: str, datatype: type):
    return {
        "type": "string",
        "format": "binary"
    }


def _builtin_schema(type_name: str, type_format: Optional[str],
                    event_name: str, datatype: type) -> dict:
    """
    Build type schema for predefined datatypes
    """
    schema = {
        "type": "object",
        "required": [
            event_name
        ],
        "properties": {
            event_name: {
                "type": type_name,
            }
        },
        "description": f"{event_name} {type_name} payload"
    }
    if type_format is not None:
        schema['properties'][event_name]['format'] = type_format  # type: ignore
    return schema


TYPE_MAPPERS = {
    str: partial(_builtin_schema, 'string', None),
    int: partial(_builtin_schema, 'integer', None),
    float: partial(_builtin_schema, 'number', None),
    bool: partial(_builtin_schema, 'boolean', None),
    list: _array_schema,
    BinaryAttachment: partial(_builtin_schema, 'string', 'binary'),
    BinaryDownload: _binary_download_schema
}

BUILTIN_TYPES = {
    str: ('string', None),
    int: ('integer', None),
    float: ('number', None),
    bool: ('boolean', None),
    date: ('string', 'date'),
    datetime: ('string', 'date-time')
}



"""
Utilities to import apps modules and datatypes at runtime / server start
"""
from importlib import import_module
from types import ModuleType
from typing import Type

from hopeit.app.config import AppConfig, EventDescriptor
from hopeit.server.names import module_name

from hopeit.dataobjects import DataObject

__all__ = ['find_event_handler']


def find_event_handler(*, app_config: AppConfig,
                       event_name: str,
                       event_info: EventDescriptor) -> ModuleType:
    """
    Returns the initialized module implementing the event business logic.
    """
    if event_info.impl:
        return import_module(event_info.impl)

    imps = (
        app_config.engine.import_modules
        if app_config.engine and app_config.engine.import_modules
        else [app_config.app.name]
    )

    errors = []
    for imp in imps:
        try:
            qualified_name = module_name(imp, event_name)
            return import_module(qualified_name)
        except ImportError as e:
            errors.append(e)
    raise ImportError(f"Cannot import {event_name} from {imps}", *errors)


def find_datobject_type(qual_type_name: str) -> Type[DataObject]:
    mod_name,  type_name = (
        '.'.join(qual_type_name.split('.')[:-1]),
        qual_type_name.split('.')[-1]
    )
    module = import_module(mod_name)
    datatype = getattr(module, type_name)
    assert hasattr(datatype, "__data_object__"), \
        f"Type {qual_type_name} must be annotated with `@dataobject`."
    return datatype



"""
Mock hooks for testing apps
"""
from typing import Optional, AsyncGenerator, Dict, AsyncIterator, Union

from multidict import CIMultiDict


class MockField:
    """
    Replacement for field object to be used for testing form preprocess methods
    """
    def __init__(self, name: str, value: Union[str, dict], file_data: Optional[bytes] = None,
                 headers: Optional[CIMultiDict] = None) -> None:
        self.name = name
        self._value = value
        self.filename = None if file_data is None else value
        self.file_data = file_data
        self.headers = CIMultiDict({
            'Content-Type': 'application/json' if isinstance(value, dict)
            else 'application/octect-stream' if file_data is not None
            else 'text/plain'
        })

    async def text(self):
        if isinstance(self._value, str):
            return self._value
        raise TypeError("field value must be str")

    async def json(self):
        if isinstance(self._value, dict):
            return self._value
        raise TypeError("field value must be dict")

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name}, value={self._value}, filename={self.filename}, " \
               f"file_data={None if self.file_data is None else self.file_data[0:min(len(self.file_data),10)]}...)"


class MockFileHook:
    """
    Replacement for multipart file upload
    """
    def __init__(self, *, name: str, file_name: str, data: MockField):
        self.name = name
        self.file_name = file_name
        self.file_data = data.file_data
        self.size = 0

    async def read_chunks(self, *, chunk_size: int = 16) -> AsyncGenerator[bytes, None]:
        if self.file_data is None:
            return
        for i in range(0, len(self.file_data), chunk_size):
            yield self.file_data[i:min(len(self.file_data), i+chunk_size)]


class MockMultipartReader:
    """
    Replacement for Multipart reader
    """
    def __init__(self, fields: Dict[str, str], attachments: Dict[str, bytes]):
        self.fields = fields
        self.it = iter([
            MockField(name=k, value=v, file_data=attachments.get(k))
            for k, v in fields.items()
        ])

    def __aiter__(self) -> AsyncIterator[MockField]:
        return self  # type: ignore

    async def __anext__(self) -> MockField:
        try:
            part = next(self.it)
        except StopIteration:
            raise StopAsyncIteration  # pylint: disable=raise-missing-from
        return part

    async def next(self) -> MockField:
        return next(self.it)

    async def read_chunk(self, size: int) -> bytes:
        raise NotImplementedError()



"""
Test utilities/wrappers for app development
"""
from pathlib import Path
from types import ModuleType
from typing import Union, Callable, List, Optional, Dict, Tuple, Any
from datetime import datetime, timezone

from hopeit.app.config import AppConfig, parse_app_config_json, EventDescriptor
from hopeit.app.context import EventContext, PostprocessHook, PreprocessHook
from hopeit.dataobjects import DataObject, EventPayload
from hopeit.dataobjects.payload import Payload
from hopeit.server.config import AuthType, ServerConfig, LoggingConfig
from hopeit.server.events import EventHandler, get_event_settings
from hopeit.server.steps import split_event_stages, find_datatype_handler
from hopeit.server.imports import find_event_handler
from hopeit.server.logger import engine_logger
from hopeit.testing.hooks import MockFileHook, MockMultipartReader
from multidict import CIMultiDict, CIMultiDictProxy

__all__ = [
    'config',
    'server_config',
    'create_test_context',
    'execute_event'
]

logger = engine_logger()


def config(path: Union[str, Path]) -> AppConfig:
    if isinstance(path, str):
        path = Path(path)
    with open(path, 'r', encoding="utf-8") as f:
        app_config = parse_app_config_json(f.read())
        app_config.server = server_config()
        return app_config


def server_config():
    return ServerConfig(
        logging=LoggingConfig(log_level='DEBUG', log_path='logs/')
    )


def create_test_context(app_config: AppConfig, event_name: str,
                        track_ids: Optional[dict] = None,
                        auth_info: Optional[dict] = None) -> EventContext:
    """
    Creates an EventContext object to be used in tests

    :param app_config: AppConfig, app confioguration
    :param event_name: str, event_name to be called
    :param track_ids: dict, optional: addictional key/values to add to track_ids section. By default
        required track_ids will be generated with default test or empty values.
    :param auth_info: dict, optional: additional auth_info to inject. By default Unsecured auth_info
        will be generated.
    """
    return EventContext(
        app_config=app_config,
        plugin_config=app_config,
        event_name=event_name,
        settings=get_event_settings(app_config.effective_settings, event_name),  # type: ignore
        track_ids={
            **{
                k: '' for k in app_config.engine.track_headers
            },
            'track.operation_id': 'test_operation_id',
            'track.request_id': 'test_request_id',
            'track.request_ts': datetime.now(tz=timezone.utc).isoformat(),
            **({} if track_ids is None else track_ids)
        },
        auth_info={
            'auth_type': AuthType.UNSECURED, 'allowed': 'true',
            **({} if auth_info is None else auth_info)
        }
    )


def _apply_mocks(context: EventContext,
                 handler: EventHandler,
                 event_name: str,
                 effective_events: Dict[str, EventDescriptor],
                 preprocess_hook: Optional[PreprocessHook],
                 postprocess_hook: Optional[PostprocessHook],
                 mocks: List[Callable[[ModuleType, EventContext], None]]):
    """
    Execute a list of functions to mock module properties.
    """
    module, _, _ = handler.modules[event_name]
    logger.debug(context, f"[test.apps] executing mocks for module={module.__name__}...")
    for mock in mocks:
        hooks: Dict[str, Any] = {}
        if preprocess_hook is not None:
            hooks['preprocess_hook'] = preprocess_hook
        if postprocess_hook is not None:
            hooks['postprocess_hook'] = postprocess_hook
        mock(module, context, **hooks)  # type: ignore
    handler.load_modules(effective_events=effective_events)
    logger.debug(context, '[test.apps] mocking done.')


async def execute_event(app_config: AppConfig,
                        event_name: str,
                        payload: Optional[EventPayload],
                        mocks: Optional[List[Callable[[ModuleType, EventContext], None]]] = None,
                        *,
                        fields: Optional[Dict[str, str]] = None,
                        upload: Optional[Dict[str, bytes]] = None,
                        preprocess: bool = False,
                        postprocess: bool = False,
                        context: Optional[EventContext] = None,
                        **kwargs) -> Union[
                            Optional[EventPayload],
                            List[EventPayload],
                            Tuple[Optional[EventPayload], EventPayload, PostprocessHook],
                            Tuple[List[EventPayload], EventPayload, PostprocessHook]]:
    """
    Test executes an app event.

    Notice that event implementation file needs to be saved to disk since this will simulate
    execution similar to how engine actually execute events. Writing to stream will be ignored.

    :param app_config: AppConfig, load using `app_config = config('path/to/app-config.json')`
    :param event_name: str, name of the event / module to execute
    :param payload: test payload to send to initial step
    :param mocks: lists of functions to execute in order to mock functionality
    :param postprocess: enables testing __postprocess__ called with last step result or
        result before a SHUFFLE step if present.
    :param context: EventContext, optional EventContext to use when calling event. If not provided
        a default context will be created.
    :param kwargs: that will be forwarded to the initial step of the event
    :return: the results of executing the event, for simple events it will be a single object,
        for events with initial Spawn[...] the results will be collected as a list.
        If postprocess is true, a tuple of 3 elements is return, first element is results as described
        above, second element the output of call to __postprocess__, and third one a PostprocessHook
        with response information used during call to __postprocess__
    """
    async def _postprocess(hook: PostprocessHook, results: List[EventPayload]) -> EventPayload:
        assert context is not None
        pp_payload = results[-1] if len(results) > 0 else None
        return await handler.postprocess(context=context, payload=pp_payload, response=hook)

    async def _preprocess(hook: PreprocessHook, payload: EventPayload) -> EventPayload:
        assert context is not None
        return await handler.preprocess(
            context=context, query_args=kwargs, payload=payload, request=hook)

    if context is None:
        context = create_test_context(app_config, event_name)

    event_info = app_config.events[event_name]
    impl = find_event_handler(app_config=app_config, event_name=event_name, event_info=event_info)
    effective_events = {**split_event_stages(app_config.app, event_name, event_info, impl)}
    handler = EventHandler(
        app_config=app_config, plugins=[],
        effective_events=effective_events,
        settings=app_config.effective_settings  # type: ignore
    )

    preprocess_hook, postprocess_hook = None, None
    if preprocess:
        preprocess_hook = PreprocessHook(
            headers=CIMultiDictProxy(CIMultiDict()),
            multipart_reader=MockMultipartReader(fields or {}, upload or {}),  # type: ignore
            file_hook_factory=MockFileHook,
            payload_raw=b'' if payload is None else Payload.to_json(payload).encode()
        )
    if postprocess:
        postprocess_hook = PostprocessHook()
    if mocks is not None:
        _apply_mocks(context, handler, event_name, effective_events, preprocess_hook, postprocess_hook, mocks)

    datatype = find_datatype_handler(app_config=app_config, event_name=event_name, event_info=event_info)
    if preprocess_hook:
        payload = await _preprocess(preprocess_hook, payload)
        if postprocess_hook and preprocess_hook.status is not None:
            postprocess_hook.set_status(preprocess_hook.status)
    elif datatype is None:
        if payload is not None:
            return (payload, payload, postprocess_hook) if postprocess else payload
    elif not (datatype is DataObject or isinstance(payload, datatype)):
        return (payload, payload, postprocess_hook) if postprocess else payload

    on_queue, pp_result, pp_called = [payload], None, False
    for effective_event_name, event_info in effective_events.items():
        context = create_test_context(
            app_config, effective_event_name, track_ids=context.track_ids, auth_info=context.auth_info
        )
        stage_results = []
        for elem in on_queue:
            async for res in handler.handle_async_event(context=context, query_args=kwargs, payload=elem):
                stage_results.append(res)
        on_queue = stage_results if len(stage_results) > 0 else on_queue
        if postprocess_hook and not pp_called:
            pp_called = True
            pp_result = await _postprocess(postprocess_hook, on_queue)
        kwargs = {}

    if postprocess:
        if len(on_queue) == 0:
            return None, pp_result, postprocess_hook
        if len(on_queue) == 1:
            return on_queue[0], pp_result, postprocess_hook
        return list(on_queue), pp_result, postprocess_hook

    if len(on_queue) == 0:
        return None
    if len(on_queue) == 1:
        return on_queue[0]
    return list(on_queue)


async def execute_service(app_config: AppConfig,
                          event_name: str,
                          max_events: int = 1,
                          mocks: Optional[List[Callable[[ModuleType, EventContext], None]]] = None) \
        -> List[Union[EventPayload, Exception]]:
    """
    Executes __service__ handler of an event,
    and processes a maximum of `max_events`.
    :param app_config: AppConfig
    :param event_name: event_name, for a SERVICE event
    :param max_events: int, default 1, number of events to process as part of the test
    :param mocks: mocks to be forwarded to `execute_event`
    :return: List of results of processing available events
    """
    context = create_test_context(app_config, event_name)
    event_info = app_config.events[event_name]
    impl = find_event_handler(app_config=app_config, event_name=event_name, event_info=event_info)
    handler = getattr(impl, '__service__')
    count = 0
    results = []
    async for payload in handler(context):
        results.append(await execute_event(app_config, event_name, payload, mocks))
        count += 1
        if count >= max_events:
            break
    return results



"""
hopeit.engine app modules

Overview:
    * **api**: provides helpers to define event Open API compliant events APIs
    * **config**: provides schema and load helpers for application config json files
    * **context**: provides EventContext object for App Events
    * **events**: provides helpers to enhace events behaviour, like Spawan and SHUFFLE
    * **logger**: provides logging for App Events.
    * **errors**: provides error handling for app events
"""
__all__ = ['api',
           'config',
           'context',
           'errors',
           'events',
           'logger']



"""
Base class and helper functions to defined and invoke external apps
using clients plugins.
"""
from typing import Optional, Type, List
from abc import ABC
from importlib import import_module

from hopeit.app.context import EventContext
from hopeit.app.config import AppConfig
from hopeit.dataobjects import EventPayload, EventPayloadType
from hopeit.server.logger import engine_logger, extra_logger

logger = engine_logger()
extra = extra_logger()

_registered_clients = {}


class ClientException(Exception):
    """Base exception for Client errors"""


class AppConnectionNotFound(ClientException):
    """Invalid app_connection or not registered"""


class Client(ABC):
    """
    Base class to imeplement stream management of a Hopeit App
    """
    @staticmethod
    def create(app_config: AppConfig, app_connection: str):
        connection_info = app_config.app_connections[app_connection]
        cm_comps = connection_info.client.split('.')
        module_name, impl_name = '.'.join(cm_comps[:-1]), cm_comps[-1]
        logger.info(__name__, f"Importing Client module: {module_name} implementation: {impl_name}...")
        module = import_module(module_name)
        impl = getattr(module, impl_name)
        logger.info(__name__, f"Creating {impl_name} for app_connection {app_connection}...")
        return impl(app_config, app_connection)

    async def start(self):
        """
        Starts the client, create connections if necessary
        """
        raise NotImplementedError()

    async def stop(self):
        """
        Stops the client, release connections if needed
        """
        raise NotImplementedError()

    async def call(self, event_name: str,
                   *, datatype: Type[EventPayloadType], payload: Optional[EventPayload],
                   context: EventContext, **kwargs) -> List[EventPayloadType]:
        """
        Implement invocation to external apps in the configured app_connection.

        Notice that this method must always returns a list. In case the target event returns a single
        item, make this method to return a list of one element. This is to ensure type-checks.

        This method is not usually called directly, use instead `app_call` and `app_call_list` functions.

        :param event: str, event name to invoke in external app, must be configured in event connections section
        :param datatype: str, type of items returned
        :param payload: payload to pass to taget event
        :param context: current EventContext
        :param **kwargs: query args to be passed to target event

        :return: list of items of datatype
        """
        raise NotImplementedError()


async def register_app_connections(app_config: AppConfig):
    """
    Used by the engine to initialize and keep client connections
    """
    _registered_clients[app_config.app_key()] = {
        app_connection: await Client.create(app_config, app_connection).start()
        for app_connection in app_config.app_connections.keys()
    }


async def stop_app_connections(app_key: str):
    """
    Used by the engine to signal client stop and close connections
    """
    for _, client in _registered_clients[app_key].items():
        await client.stop()
    del _registered_clients[app_key]


def app_client(app_connection: str, context: EventContext) -> Client:
    """
    Returns the cached client for a given app_connection
    """
    try:
        return _registered_clients[context.app_key][app_connection]
    except KeyError:
        raise AppConnectionNotFound(  # pylint: disable=raise-missing-from
            f"Not found app_connection: {app_connection} for app: {context.app_key}"
        )


async def app_call(app_connection: str,
                   *, event: str, datatype: Type[EventPayloadType],
                   payload: EventPayload, context: EventContext,
                   **kwargs) -> EventPayloadType:
    """
    Invokes event in external app using configured app_connection, for events that return
    a single item.

    :param app_connection: str, app_connection name as in app_config
    :param event: str, event name to invoke in external app, must be configured in event connections section
    :param datatype: str, type of returned value
    :param payload: payload to pass to taget event
    :param context: current EventContext
    :param **kwargs: query args to be passed to target event
    """
    client = app_client(app_connection, context)
    results = await client.call(
        event, datatype=datatype, payload=payload, context=context, **kwargs
    )
    return results[0]


async def app_call_list(app_connection: str,
                        *, event: str, datatype: Type[EventPayloadType],
                        payload: EventPayload, context: EventContext,
                        **kwargs) -> List[EventPayloadType]:
    """
    Invokes event in external app using configured app_connection, for events that return
    a list or collection of items.

    :param app_connection: str, app_connection name as in app_config
    :param event: str, event name to invoke in external app, must be configured in event connections section
    :param datatype: str, type of returned value for each item
    :param payload: payload to pass to taget event
    :param context: current EventContext
    :param kwargs: query args to be passed to target event
    """
    client = app_client(app_connection, context)
    return await client.call(
        event, datatype=datatype, payload=payload, context=context, **kwargs
    )



"""
Server configuration utility
"""
from enum import Enum
from typing import TypeVar, List, Optional
import re
import os

from dataclasses import dataclass, field

from hopeit.dataobjects import dataobject
from hopeit.server.names import auto_path_prefixed
from hopeit.server.version import ENGINE_VERSION


__all__ = ['StreamsConfig',
           'LoggingConfig',
           'AuthType',
           'AuthConfig',
           'ServerConfig',
           'parse_server_config_json',
           'replace_env_vars',
           'replace_config_args']


DEFAULT_STR = "<<DEFAULT>>"
ConfigType = TypeVar("ConfigType")  # pylint: disable=invalid-name


@dataobject
@dataclass
class StreamsConfig:
    """
    :field connection_str: str, url to connect to streams server: i.e. redis://localhost:6379
    """
    stream_manager: str = "hopeit.streams.NoStreamManager"
    connection_str: str = '<<NoStreamManager>>'
    delay_auto_start_seconds: int = 3


@dataobject
@dataclass
class LoggingConfig:
    log_level: str = 'INFO'
    log_path: str = 'logs/'


class AuthType(str, Enum):
    """
    Supported Authorization/Authentication types
    """
    UNSECURED = 'Unsecured'
    BASIC = 'Basic'
    BEARER = 'Bearer'
    REFRESH = 'Refresh'


@dataobject
@dataclass
class AuthConfig:
    """
    Server configuration to handle authorization tokens
    """
    secrets_location: str
    auth_passphrase: str
    enabled: bool = True
    create_keys: bool = False
    domain: Optional[str] = None
    encryption_algorithm: str = 'RS256'
    default_auth_methods: List[AuthType] = field(default_factory=list)

    def __post_init__(self):
        if len(self.default_auth_methods) == 0:
            self.default_auth_methods.append(AuthType.UNSECURED)

    @staticmethod
    def no_auth():
        return AuthConfig('.secrets/', '', enabled=False)


@dataobject
@dataclass
class APIConfig:
    """
    Config for Open API docs page
    """
    docs_path: Optional[str] = None


@dataobject
@dataclass
class ServerConfig:
    """
    Server configuration
    """
    streams: StreamsConfig = field(default_factory=StreamsConfig)
    logging: LoggingConfig = field(default_factory=LoggingConfig)
    auth: AuthConfig = field(default_factory=AuthConfig.no_auth)
    api: APIConfig = field(default_factory=APIConfig)
    engine_version: str = field(default=ENGINE_VERSION)

    def __post_init__(self):
        self.engine_version = ENGINE_VERSION


def parse_server_config_json(config_json: str) -> ServerConfig:
    """
    Parses configuration json file contents into EngineConfig data structure.
    Before conversion, parameters enclosed with { } are replaced by its
    respective values (@see _replace_args)
    """
    effective_json = replace_env_vars(config_json)
    parsed_config = ServerConfig.from_json(effective_json)  # type: ignore
    replace_config_args(
        parsed_config=parsed_config,
        config_classes=tuple([StreamsConfig])
    )
    return parsed_config


def replace_env_vars(config_json: str) -> str:
    """
    Replaces env variables matching form ${VAR_NAME} with its string value
    :param config_json: input configuratio json as string
    :return: str, with replaced values
    :raise: AssertionError if variables matching ${VAR_NAME} form are not replaced
    """
    result = config_json
    env_re = re.compile('\\${([^}{]+)}', re.IGNORECASE)
    for match in env_re.finditer(result):
        expr = match.group(0)
        var_name = match.group(1)
        value = os.getenv(var_name.upper())
        if value:
            result = result.replace(expr, value)

    missing_env_vars = env_re.findall(result)
    assert len(missing_env_vars) == 0, \
        f"Cannot get value from OS environment vars: {missing_env_vars}"

    return result


def replace_config_args(*, parsed_config: ConfigType,
                        config_classes: tuple,
                        auto_prefix: str = ''):
    """
    Replaces {...} enclosed expression in string values inside parsed_config.
    {...} expressions are paths to objects in same parsed_config object, expressed
    using dot notation. Also special name `{auto}` is replaced by the object path.
    parsed_config is modified inline and no value is returned.
    Replacement is done twice so values that refers to other generated values
    are replaced.

    Example:
    `{app.name}` will be replaced by the contents of `parsed_config.app.name`

    `{auto}` will be replaced by the path of keys to the current object
    built using dot notation and prefixed by `app.name`.`app.version`

    Example::

        AppConfig(
            app=App(name="myapp", version="1x0"),
            events={"event1": EventDescriptor(
                steps={"step1": StepDescriptor(notify="{auto}" ... )}
            )}
        )

    `{auto}` will be replaced by `myapp.1x0.event1.step1`
    """

    def _replace_in_dict(node: dict, expr: str, replacement: str):
        key_replacements = {}
        for k, value in node.items():
            if expr in k:
                key_replacements[k] = k.replace(expr, replacement)
            if isinstance(value, str):
                if expr in value:
                    node[k] = value.replace(expr, replacement)
            elif isinstance(value, (dict, list, *config_classes)):
                _replace_in_config(value, expr, replacement)
        for k, r in key_replacements.items():
            node[r] = node[k]
            del node[k]

    def _replace_in_list(node: list, expr: str, replacement: str):
        for i, value in enumerate(node):
            if isinstance(value, str):
                if expr in value:
                    node[i] = value.replace(expr, replacement)
            elif isinstance(value, (dict, list, *config_classes)):
                _replace_in_config(value, expr, replacement)

    def _replace_in_config(node, expr: str, replacement: str):
        if isinstance(node, dict):
            _replace_in_dict(node, expr, replacement)
        if isinstance(node, list):
            _replace_in_list(node, expr, replacement)
        else:
            for attr_name in dir(node):
                if attr_name[0] != '_' and hasattr(node, attr_name):
                    value = getattr(node, attr_name)
                    if isinstance(value, str):
                        if expr in value:
                            value = value.replace(expr, replacement)
                            setattr(node, attr_name, value)
                    elif isinstance(value, (dict, list, *config_classes)):
                        _replace_in_config(value, expr, replacement)

    def _replace_dict_items(*, node, prefix: str, this_path_prefix: str):
        for attr_name, value in node.items():
            prefix_attr = f"{prefix}.{attr_name}".lstrip('.')
            this_path = auto_path_prefixed(this_path_prefix, *attr_name.split('.'))
            if isinstance(value, str):
                if '{auto}' in value:
                    node[attr_name] = value.replace('{auto}', this_path)
                else:
                    expr = '{' + prefix_attr + '}'
                    _replace_in_config(parsed_config, expr, value)
            elif isinstance(value, dict):
                _replace_dict_items(node=value, prefix=prefix_attr, this_path_prefix=this_path)
            elif isinstance(value, list):
                _replace_list_items(node=value, prefix=prefix_attr, this_path_prefix=this_path)
            elif isinstance(value, config_classes):
                _replace_attrs(node=value, prefix=prefix_attr, this_path_prefix=this_path)

    def _replace_list_items(*, node, prefix: str, this_path_prefix: str):
        for value in node:
            if isinstance(value, dict):
                _replace_dict_items(node=value, prefix=prefix, this_path_prefix=this_path_prefix)
            if isinstance(value, list):
                _replace_list_items(node=value, prefix=prefix, this_path_prefix=this_path_prefix)
            elif isinstance(value, config_classes):
                _replace_attrs(node=value, prefix=prefix, this_path_prefix=this_path_prefix)

    def _replace_attrs(*, node, prefix: str, this_path_prefix: str):
        for attr_name in dir(node):
            if attr_name[0] != '_' and hasattr(node, attr_name):
                prefix_attr = f"{prefix}.{attr_name}".lstrip('.')
                this_path = this_path_prefix
                value = getattr(node, attr_name)
                if isinstance(value, str):
                    if '{auto}' in value:
                        setattr(node, attr_name, value.replace('{auto}', this_path))
                    else:
                        expr = '{' + prefix_attr + '}'
                        _replace_in_config(parsed_config, expr, value)
                elif isinstance(value, dict):
                    _replace_dict_items(node=value, prefix=prefix_attr, this_path_prefix=this_path)
                elif isinstance(value, list):
                    _replace_list_items(node=value, prefix=prefix_attr, this_path_prefix=this_path)
                elif isinstance(value, config_classes):
                    _replace_attrs(node=value, prefix=prefix_attr, this_path_prefix=this_path)

    for _ in range(2):
        _replace_attrs(
            node=parsed_config,
            prefix='.',
            this_path_prefix=auto_prefix
        )



"""
CLI server commands
"""
import sys

try:
    import click
    from hopeit.server import web
except ModuleNotFoundError:
    print("ERROR: Missing dependencies."
          "\n       To use hopeit_server command line tool"
          "\n       install using `pip install hopeit.engine[web,cli]`")
    sys.exit(1)


@click.group()
def server():
    pass


@server.command()
@click.option('--config-files', prompt='Config files',
              help='Comma-separated config file paths, starting with server config, then plugins, then apps.')
@click.option('--api-file', help='Path to openapi complaint json specification.')
@click.option('--host', default='0.0.0.0', help='Server host address or name.')
@click.option('--port', default='8020', help='TCP/IP port to listen.')
@click.option('--path', help='POSIX complaint socket name.')
@click.option('--start-streams', is_flag=True, default=False, help='Auto start reading stream events.')
@click.option('--enabled-groups', default='',
              help="Optional comma-separated group labels to start. If no group is specified, all events will be"
              " started. Events with no group or 'DEFAULT' group label will always be started. 'DEFAULT' group label"
              " can also be used explicitly to start only events with no group or 'DEFAULT' group label.")
def run(config_files: str, api_file: str, host: str, port: int, path: str, start_streams: bool, enabled_groups: str):
    """
    Runs web server hosting apps specified in config files.
    """
    web.prepare_engine(
        config_files=config_files.split(','),
        api_file=api_file,
        enabled_groups=enabled_groups.split(',') if enabled_groups else [],
        start_streams=start_streams,
    )
    web.serve(host=host, path=path, port=port)


cli = click.CommandCollection(sources=[server])

if __name__ == '__main__':
    cli()



"""
Convenience methods to normalize endpoint names, module names from configuration
"""
import re

from stringcase import spinalcase, snakecase  # type: ignore

__all__ = ['route_name',
           'module_name',
           'auto_path',
           'auto_path_prefixed']


def route_name(*args: str) -> str:
    return '/' + '/'.join(spinalcase(re.sub('\\.', 'x', name)) for name in args)


def module_name(*args: str) -> str:
    return '.'.join(
        re.sub('\\$', '.', snakecase(re.sub('\\.', '$', name)))
        for name in args)


def auto_path(*args: str) -> str:
    return '.'.join(snakecase(re.sub('\\.', 'x', name)) for name in args)


def auto_path_prefixed(prefix, *args: str) -> str:
    return '.'.join([
        prefix,
        *(snakecase(re.sub('\\.', 'x', name)) for name in args)
    ])



"""
hopeit.engine CLI (Command Line Interface) module

provides implementation for the following CLI commands:

    * **openapi** (hopeit_openapi): creation, diff and update openapi.json spec files
    * **server** (hopeit_server): tool for running a server instance
"""



"""
Engine version constants.
Increment on release
To ensure configuration files from example apps and plugins have same version as engine,
environment variables `HOPEIT_ENGINE_VERSION` and `HOPEIT_APPS_API_VERSION` are set.
"""
import os
import sys

ENGINE_NAME = "hopeit.engine"
ENGINE_VERSION = "0.16.3"

# Major.Minor version to be used in App versions and Api endpoints for Apps/Plugins
APPS_API_VERSION = '.'.join(ENGINE_VERSION.split('.')[0:2])
APPS_ROUTE_VERSION = APPS_API_VERSION.replace('.', 'x')

os.environ['HOPEIT_ENGINE_VERSION'] = ENGINE_VERSION
os.environ['HOPEIT_APPS_API_VERSION'] = APPS_API_VERSION
os.environ['HOPEIT_APPS_ROUTE_VERSION'] = APPS_ROUTE_VERSION

if __name__ == "__main__":
    if len(sys.argv) > 1 and sys.argv[1] == "APPS_API_VERSION":
        print(APPS_API_VERSION)
    elif len(sys.argv) > 1 and sys.argv[1] == "APPS_ROUTE_VERSION":
        print(APPS_ROUTE_VERSION)
    else:
        print(ENGINE_VERSION)



"""
hopeit.engine toolkit

Helper modules intended to be used by apps to solve common problems

    * **auth**: helpers to work with authentication and authorization mechanisms
"""

__all__ = ['auth']



"""
Utilities for handling Web requests
"""
from typing import Union
from pathlib import Path

from hopeit.app.context import PreprocessFileHook


async def save_multipart_attachment(file_hook: PreprocessFileHook, path: Union[str, Path],
                                    *, chunk_size: int = 8192):
    """
    Save file using chunks from multipart upload to specified path
    """
    with open(path, 'wb') as f:
        async for chunk in file_hook.read_chunks(chunk_size=chunk_size):
            f.write(chunk)



"""
Server/Engine logging module
"""
import logging
import os
import socket
import time
from datetime import datetime
from functools import partial
from logging.handlers import WatchedFileHandler
from typing import Dict, Iterable, Union, List, Tuple, Callable, Any
from stringcase import snakecase  # type: ignore

from hopeit.server import version
from hopeit.app.config import AppDescriptor, AppConfig, EventSettings
from hopeit.app.context import EventContext
from hopeit.server.errors import json_exc
from hopeit.server.config import ServerConfig, LoggingConfig

DEFAULT_ENGINE_LOGGER = 'engine_logger_default'
LOG_FORMAT = "%(asctime)s | %(levelname)s | %(name)s | %(message)s | %(extra)s"
WARNINGS_LOG_FORMAT = "%(asctime)s | %(levelname)s | %(name)s | %(message)s | "

__all__ = ['extra_values',
           'combined',
           'setup_app_logger',
           'engine_logger',
           'extra_logger',
           'format_extra_values',
           'EngineLoggerWrapper']


class EventLoggerWrapper:
    """
    Wraps standard python logger to include request_id on every entry.
    """

    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def debug(self, context: EventContext, msg, *args, **kwargs) -> None:
        if self.logger.isEnabledFor(logging.DEBUG):
            _enrich_extra(kwargs, context, msg)
            self.logger.debug(msg, *args, **kwargs)

    def info(self, context: EventContext, msg, *args, **kwargs) -> None:
        _enrich_extra(kwargs, context, msg)
        self.logger.info(msg, *args, **kwargs)

    def warning(self, context: EventContext, msg, *args, **kwargs) -> None:
        _enrich_extra(kwargs, context, msg)
        self.logger.warning(msg, *args, **kwargs)

    def error(self, context: EventContext, msg, *args, **kwargs) -> None:
        _enrich_extra(kwargs, context, msg)
        self.logger.error(msg, *args, **kwargs)


def _logger_name(app: AppDescriptor, name: str):
    host = socket.gethostname()
    pid = os.getpid()
    base_name = name.split('$')[0]
    return f"{app.name} {app.version} {base_name} {host} {pid}"


def _engine_logger_name(name: str):
    host = socket.gethostname()
    pid = os.getpid()
    return f"{version.ENGINE_NAME} {version.ENGINE_VERSION} {name} {host} {pid}"


def extra_values(required_fields: Iterable[str], *, prefix='extra.', **kwargs) -> Dict[str, str]:
    values = {
        **{k: kwargs[k] for k in required_fields},
        **kwargs
    }
    return {'extra': format_extra_values(values, prefix=prefix)}


def extra_logger():
    return partial(extra_values, [])


def combined(*args) -> Dict[str, str]:
    return {'extra': ' | '.join(x['extra'] for x in args)}


def format_extra_values(values: Dict[str, Any], prefix: str = '') -> str:
    return ' | '.join(
        f"{prefix}{k}={_format_value(v)}" for k, v in values.items()
    )


def _format_value(v: Any) -> str:
    if isinstance(v, float):
        return f"{v:.3f}"
    if isinstance(v, datetime):
        return v.isoformat()
    return str(v)


def _enrich_extra(logger_kwargs, context: Union[str, EventContext], msg=None):
    logger_kwargs['extra'] = logger_kwargs.get('extra', {'extra': ''})
    if isinstance(context, EventContext):
        logger_kwargs['extra']['extra'] += f" | {format_extra_values(context.track_ids)}"
    if isinstance(msg, Exception):
        logger_kwargs['extra']['extra'] += \
            f" | trace={json_exc(msg)}"
    logger_kwargs['extra']['extra'] = logger_kwargs['extra']['extra'].lstrip('| ')


def setup_extra_fields_extractor(module, *, event_settings: EventSettings):
    extra_fields = event_settings.logging.extra_fields
    module.extra = partial(extra_values, extra_fields)


def _file_handler(logger_name: str, formatter: logging.Formatter, log_path: str):
    os.makedirs(log_path, exist_ok=True)
    path = f"{log_path}{snakecase(logger_name.replace('.', 'x'))}.log"
    file_handler = WatchedFileHandler(path)
    file_handler.setFormatter(formatter)
    return file_handler


def _console_handler(logger_name: str, formatter: logging.Formatter):
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    return ch


def _setup_standard_logger(logger_name: str, config: LoggingConfig, log_format: str = LOG_FORMAT) -> logging.Logger:
    """
    Creates a python logger using server logging configuration.
    By default all loggers are created with a file handler.
    If log_leve is DEBUG, an additional console handler is attached.

    :param logger_name: logger name
    :param config: server logging config
    :param log_format: format string, default to LOG_FORMAT
    :return: python standard logger
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(config.log_level)
    formatter = logging.Formatter(log_format)
    formatter.converter = time.gmtime  # type: ignore
    file_handler = _file_handler(logger_name, formatter, config.log_path)
    logger.addHandler(file_handler)
    if config.log_level == "DEBUG":
        ch = _console_handler(logger_name, formatter)
        logger.addHandler(ch)
    return logger


def _setup_console_logger(logger_name: str) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(LOG_FORMAT)
    formatter.converter = time.gmtime  # type: ignore
    ch = _console_handler(logger_name, formatter)
    logger.addHandler(ch)
    return logger


class EngineLoggerWrapper:
    """
    Wrapper around standard python Logger to be used in engine modules.
    Provides additional functionallity to log extra info and metrics.
    """
    engine_logger: logging.Logger = logging.getLogger("bootstrap")
    loggers: Dict[str, logging.Logger] = {}

    def init_cli(self, name: str):
        EngineLoggerWrapper.engine_logger = _setup_console_logger(_engine_logger_name(name))
        return self

    def init_server(self, server_config: ServerConfig):
        _setup_standard_logger("py.warnings", config=server_config.logging, log_format=WARNINGS_LOG_FORMAT)
        logging.captureWarnings(True)
        EngineLoggerWrapper.engine_logger = _setup_standard_logger(
            _engine_logger_name("engine"), config=server_config.logging)
        return self

    def init_app(self, app_config: AppConfig, plugins: List[AppConfig]):
        """
        Initializes logger for an app and its plugins
        """
        assert app_config.server
        events = [
            *app_config.events.keys(),
            *[k for plugin in plugins for k in plugin.events.keys()]
        ]
        for event_name in events:
            logger_name = _logger_name(app_config.app, event_name)
            if logger_name not in self.loggers:
                logger = _setup_standard_logger(
                    logger_name, config=app_config.server.logging)
                self.loggers[logger_name] = logger
        return self

    def _logger(self, context: EventContext):
        logger_name = _logger_name(context.app, context.event_name)
        return self.loggers[logger_name]

    def start(self, context: EventContext, *args, **kwargs) -> None:
        self.info(context, 'START', *args, **kwargs)

    def failed(self, context: EventContext, *args, **kwargs) -> None:
        self.error(context, 'FAILED', *args, **kwargs)

    def ignored(self, context: EventContext, *args, **kwargs) -> None:
        self.warning(context, 'IGNORED', *args, **kwargs)

    def done(self, context: EventContext, *args, **kwargs) -> None:
        self.info(context, 'DONE', *args, **kwargs)

    def stats(self, context: EventContext, *args, **kwargs) -> None:
        self.info(context, 'STATS', *args, **kwargs)

    def debug(self, context: Union[str, EventContext], msg, *args, **kwargs) -> None:
        _enrich_extra(kwargs, context, msg)  # type: ignore
        if isinstance(context, str):
            self.engine_logger.debug(f"[{context}] {msg}", *args, **kwargs)
        else:
            self._logger(context).debug(msg, *args, **kwargs)  # type: ignore

    def info(self, context: Union[str, EventContext], msg, *args, **kwargs) -> None:
        _enrich_extra(kwargs, context, msg)  # type: ignore
        if isinstance(context, str):
            self.engine_logger.info(f"[{context}] {msg}", *args, **kwargs)
        else:
            self._logger(context).info(msg, *args, **kwargs)   # type: ignore

    def warning(self, context: Union[str, EventContext], msg, *args, **kwargs) -> None:
        _enrich_extra(kwargs, context, msg)  # type: ignore
        if isinstance(context, str):
            self.engine_logger.warning(f"[{context}] {msg}", *args, **kwargs)
        else:
            self._logger(context).warning(msg, *args, **kwargs)  # type: ignore

    def error(self, context: Union[str, EventContext], msg, *args, **kwargs) -> None:
        _enrich_extra(kwargs, context, msg)  # type: ignore
        if isinstance(context, str):
            self.engine_logger.error(f"[{context}] {msg}", *args, **kwargs)
        else:
            self._logger(context).error(msg, *args, **kwargs)  # type: ignore


def setup_app_logger(module, *, app_config: AppConfig, name: str, event_settings: EventSettings):
    """
    Returns wrapper over python logging Logger for a given app and name.
    Logger name is made combining {app.name} {app.version} {name} {host} {pid}
    Standard fields to be logged are `%(asctime)s | %(levelname)s | %(name)s | %(message)s | `
    Specific apps can require for extra fields per event, configurabe in EventDescriptor
    """
    if hasattr(module, 'logger') and not isinstance(module.logger, EventLoggerWrapper):
        assert app_config.server
        logger_name = _logger_name(app_config.app, name)
        logger = EngineLoggerWrapper.loggers.get(logger_name)
        if logger is None:
            logger = _setup_standard_logger(logger_name, config=app_config.server.logging)
            EngineLoggerWrapper.loggers[logger_name] = logger
        module.logger = EventLoggerWrapper(logger)
        setup_extra_fields_extractor(module, event_settings=event_settings)


def engine_logger() -> EngineLoggerWrapper:
    """
    Returns logger wrapper for engine modules
    Allows to reference `logger` at module scope.

    Use at module level in events implementation::
    ```
        from hopeit.logger import engine_logger()

        logger = engine_logger()
    ```
    """
    return EngineLoggerWrapper()


def engine_extra_logger() -> Tuple[EngineLoggerWrapper, Callable]:
    """
    Returns logger wrapper for engine modules
    and a convenience function to submit extra values when logging.
    Allows to reference `logger` and `extra` at module scope.

    Use at module level in events implementation::
    ```
        from hopeit.logger import engine_logger()

        logger, extra = engine_extra_logger()

        ...

        logger.info(context, "message" extra=extra(value1="extra_value", ...))
    ```
    """
    return EngineLoggerWrapper(), partial(extra_values, [])



"""
Provides generic `serialize`, `deserialize` methods to handle payloads
"""
import base64
import pickle
from typing import Type

from hopeit.app.config import Serialization, Compression

__all__ = ['serialize', 'deserialize']

from hopeit.dataobjects import EventPayload, EventPayloadType
from hopeit.dataobjects.payload import Payload
from hopeit.server.compression import compress, decompress


def _ser_json_utf8(data: EventPayload, level: int) -> bytes:
    return Payload.to_json(data).encode('utf-8')


def _deser_json_utf8(data: bytes, datatype: Type[EventPayloadType]) -> EventPayload:
    return Payload.from_json(data.decode('utf-8'), datatype)


def _ser_pickle(data: EventPayload, level: int) -> bytes:
    return pickle.dumps(data, protocol=level)


def _deser_pickle(data: bytes, datatype: Type[EventPayloadType]) -> EventPayload:
    return pickle.loads(data)


def _ser_json_base64(data: EventPayload, level: int) -> bytes:
    return base64.b64encode(_ser_json_utf8(data, level))


def _deser_json_base64(data: bytes, datatype: Type[EventPayloadType]) -> EventPayload:
    return _deser_json_utf8(base64.b64decode(data), datatype)


_SERDESER = {
    Serialization.JSON_UTF8: (_ser_json_utf8, 0, _deser_json_utf8),
    Serialization.JSON_BASE64: (_ser_json_base64, 0, _deser_json_base64),
    Serialization.PICKLE3: (_ser_pickle, 3, _deser_pickle),
    Serialization.PICKLE4: (_ser_pickle, 4, _deser_pickle),
    Serialization.PICKLE5: (_ser_pickle, 5, _deser_pickle)
}


def serialize(data: EventPayload, serialization: Serialization, compression: Compression) -> bytes:
    algos = _SERDESER[serialization]
    encoded = algos[0](data, level=algos[1])
    return compress(encoded, compression)


def deserialize(data: bytes, serialization: Serialization, compression: Compression,
                datatype: Type[EventPayloadType]) -> EventPayload:
    algos = _SERDESER[serialization]
    decomp = decompress(data, compression)
    return algos[2](decomp, datatype)



"""
AsyncCollector (aliased as Collector) implementation to be used as a way to concurrently execute steps in an event

Use `hopeit.app.events` `collector_step(...)` constructor to define steps implementing AsyncCollector
"""
import asyncio
from typing import Callable, Dict, Any, Tuple, Optional, Coroutine
from hopeit.dataobjects import copy_payload

from hopeit.app.context import EventContext

__all__ = ['Collector', 'AsyncCollector', 'CollectorStepType']


class AbstractCollector:
    async def run(self, context: EventContext) -> Any:
        raise NotImplementedError("AbstractCollector cannot be executed.")


CollectorStepType = Callable[[AbstractCollector, EventContext], Coroutine]


class CollectorItem:
    def __init__(self, func: CollectorStepType):
        self.func = func
        self.lock = asyncio.Lock()
        self.data: Any = None


class AsyncCollector(AbstractCollector):
    """
    Allows to define a list of steps (functions) to be executed concurrently.

    Example::

        AsyncCollector.input(payload).steps(('step1', step1func, 'step2', step2func').run()


    Gathers and run concurrently step1func and step2func and put their return values in the collector
    under the specified key ('step1' or 'step2')
    If a function needs the result from another step it can be retrieved with `await collector['step1'].
    To access the payload a function can do `payload = await collector['payload']
    This way you can specify a number of functions that can run asynchronously with no particular order
    (i.e. when querying multiple databases or making requests in parallel to external services),
    and then use the results or combine them in a different step.

    CAUTION: AsyncCollector makes not guarantees whether your code will block indefinitely (i,e.
    if you do `await collector['step1']` from step2func but step1func does `await collector['step2']`. This
    will block your application, so the only way the engine will prevent this to lock your server is
    using timeouts. In case of a dead-lock, event will fail to process and concurrent functions
    will be canceled when reaching a timeout, but no checking is done on whether a deadlock is happening.
    Please check the sequence your code is accessing/awaiting results from the collector to avoid cycles.
    """
    __data_object__ = {'unsafe': True, 'validate': False, 'schema': False}

    def __init__(self):
        self.items: Dict[str, CollectorItem] = {}
        self.executed: bool = False
        self.payload: Optional[Any] = None

    def input(self, payload: Any):
        self.payload = payload
        return self

    def steps(self, *funcs: Tuple[str, CollectorStepType]):
        for name, func in funcs:
            self.items[name] = CollectorItem(func)
        return self

    async def _get(self, name, lock=False) -> Any:
        """
        Locks and waits for a collector steps is computed and return its results.
        In case name is 'payload', returns collector input without blocking.
        """
        if name == 'payload':
            return copy_payload(self.payload)
        assert self.executed, "Collector not executed. Call collector.run(...) before accessing results."
        item = self.items[name]
        await item.lock.acquire()
        try:
            return copy_payload(item.data)
        finally:
            if not lock:
                item.lock.release()

    def __getitem__(self, item):
        return self._get(item)

    async def _run_item(self, item: CollectorItem, context: EventContext):
        assert item.lock.locked(), f"Step {item} already released."
        try:
            item.data = await item.func(self, context)
        finally:
            item.lock.release()

    async def run(self, context: EventContext):
        steps = []
        for item in self.items.values():
            await item.lock.acquire()
            steps.append(self._run_item(item, context))
        self.executed = True
        await asyncio.gather(*steps)
        return self


Collector = AsyncCollector



"""
Events base classes and low level handlers to execute events specified by apps
"""
from types import ModuleType
from typing import Dict, Optional, List, Tuple, AsyncGenerator, Any
from asyncio import iscoroutine

from hopeit.dataobjects import EventPayload
from hopeit.server.steps import (
    StepExecutionList, StepInfo, extract_postprocess_handler, extract_preprocess_handler, execute_steps,
    invoke_single_step, extract_module_steps, effective_steps, event_and_step
)
from hopeit.server.logger import setup_app_logger, engine_logger, extra_logger
from hopeit.server.imports import find_event_handler
from hopeit.app.config import AppConfig, EventDescriptor, EventSettings
from hopeit.app.context import EventContext, PostprocessHook, PreprocessHook

__all__ = ['EventHandler']

logger = engine_logger()
extra = extra_logger()


class EventHandler:
    """
    Handles execution of Hopeit App events
    """

    def __init__(self, *,
                 app_config: AppConfig,
                 plugins: List[AppConfig],
                 effective_events: Dict[str, EventDescriptor],
                 settings: Dict[str, Any]):
        """
        Creates an EventHandler for a Hopeit App

        :param app_config: AppConfig, configuration for the App
        """
        self.app_config = app_config
        self.modules: Dict[str, Tuple[ModuleType, bool, list]] = {}
        self.steps: Dict[str, StepExecutionList] = {}
        self.preprocess_handlers: Dict[str, Optional[StepInfo]] = {}
        self.postprocess_handlers: Dict[str, Optional[StepInfo]] = {}
        self.settings = settings
        logger.init_app(app_config, plugins)
        self.load_modules(effective_events)

    def load_modules(self, effective_events: Dict[str, EventDescriptor]):
        """
        Load and setup event implementation modules. Set logger and load settings.
        """
        for event_name, event_info in effective_events.items():
            base_event, _ = event_and_step(event_name)
            module = find_event_handler(app_config=self.app_config, event_name=base_event, event_info=event_info)
            steps = extract_module_steps(module)
            self.modules[base_event] = (module, False, steps)
            self.preprocess_handlers[base_event] = extract_preprocess_handler(module)
            self.postprocess_handlers[base_event] = extract_postprocess_handler(module)
            event_settings = get_event_settings(self.settings, event_name)
            setup_app_logger(
                module, app_config=self.app_config, name=base_event, event_settings=event_settings
            )
            self.steps[event_name] = effective_steps(event_name, steps)

    async def _ensure_initialized(self, context: EventContext):
        base_event, _ = event_and_step(context.event_name)
        impl, initialized, raw_steps = self.modules[base_event]
        if not initialized:
            await self._init_module(module=impl, context=context)
            self.modules[base_event] = (impl, True, raw_steps)

    async def handle_async_event(self, *,
                                 context: EventContext,
                                 query_args: Optional[Dict[str, Any]],
                                 payload: Optional[EventPayload]) -> AsyncGenerator[Optional[EventPayload], None]:
        """
        Handles execution of engine defined event.
        Executes event handler code deployed with app.

        Execution goes as following:
            * EventDescriptor from AppConfig is used
            * an object from a class with CamelCase name same as event name is instantiated
            * find the next step to execute that accepts input with payload type
            * method with same name as step is invoked in instantiated object
            * if a step specifies write_stream, and event is not None, payload is published to a stream
            * repeats previous 3 steps executing next step that accepts current payload type

        :param context: EventContext
        :param query_args: arguments from a query context in the form of a dictionary
        :param payload: EventPayload, to be sent to event implementor
        """
        await self._ensure_initialized(context)
        steps = self.steps[context.event_name]
        async for result in execute_steps(
            steps, context=context,
            payload=payload, **(query_args or {})
        ):
            yield result

    async def postprocess(self, *,
                          context: EventContext,
                          payload: Optional[EventPayload],
                          response: PostprocessHook) -> Optional[EventPayload]:
        """
        Invokes postprocess method in event if defined in event configuration,
        allowing events to append headers, cookies and status to a response
        """
        pp_handler = self.postprocess_handlers[context.event_name]
        if pp_handler:
            _, initialized, _ = self.modules[context.event_name]
            assert initialized, \
                "Module not initialized. Postprocess requires events steps to be executed first"
            return await invoke_single_step(payload=payload, context=context, func=pp_handler[0], response=response)
        return payload

    async def preprocess(self, *,
                         context: EventContext,
                         query_args: Optional[Dict[str, Any]],
                         payload: Optional[EventPayload],
                         request: PreprocessHook) -> Optional[EventPayload]:
        """
        Invokes __preprocess__ method in event if defined in event,
        allowing events to process elements from requests.
        """
        pp_handler = self.preprocess_handlers[context.event_name]
        if pp_handler:
            await self._ensure_initialized(context)
            return await invoke_single_step(payload=payload, context=context,
                                            func=pp_handler[0], request=request, **(query_args or {}))
        return payload

    async def _init_module(self, *, module, context: EventContext):
        if hasattr(module, '__init_event__'):
            logger.info(context, f"__init_event__ module={module.__name__}...")
            init_f = getattr(module, '__init_event__')
            coro_or_res = init_f(context)
            if iscoroutine(coro_or_res):
                await coro_or_res  # type: ignore


def get_runtime_settings(app_config: AppConfig, plugins: List[AppConfig]) -> Dict[str, Any]:
    settings: Dict[str, Any] = {}
    for plugin in plugins:
        assert plugin.effective_settings is not None
        settings.update(plugin.effective_settings)
    assert app_config.effective_settings is not None
    settings.update(app_config.effective_settings)
    return settings


def get_event_settings(settings: Dict[str, Any], event_name: str) -> EventSettings:
    data = settings[event_name.split('$')[0]]  # Removes auto-generated event suffixes
    return EventSettings.from_dict(data, validate=False)  # type: ignore



"""
Server error handling convenience module
"""
import traceback
import json
import urllib.parse
from dataclasses import dataclass
from typing import List

from hopeit.dataobjects import dataobject

__all__ = ['ErrorInfo',
           'format_exc',
           'json_exc']


@dataobject
@dataclass
class ErrorInfo:
    """
    Error information to be returned in failed responses
    """
    msg: str
    tb: List[str]

    @staticmethod
    def from_exception(e: BaseException):
        return ErrorInfo(
            msg=str(e),
            tb=traceback.format_exception_only(type(e), e)
        )


def format_exc(e: Exception) -> List[str]:
    traceback.print_exception(None, e, e.__traceback__)
    return traceback.format_exception(None, e, e.__traceback__)


def json_exc(e: Exception) -> str:
    return urllib.parse.quote(
        json.dumps(format_exc(e))
    )
